import csv
import gzip
import logging
import re
from collections import Counter
from collections.abc import Iterable, Mapping
from copy import deepcopy
from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, Iterator, Literal, Optional, Union
from urllib.parse import quote

import httpx

if TYPE_CHECKING:
    import pandas as pd

from phoenix.client.__generated__ import v1
from phoenix.client.utils.id_handling import is_node_id

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT_IN_SECONDS = 5


def _is_valid_dataset_example(obj: Any) -> bool:
    """
    Check if an object is a valid DatasetExample using the TypedDict's annotations.
    """
    if not isinstance(obj, dict):
        return False

    required_fields = set(v1.DatasetExample.__annotations__.keys())

    if not required_fields.issubset(obj.keys()):  # pyright: ignore[reportUnknownArgumentType]
        return False
    return True


class Dataset:
    """
    A dataset with its examples and version information.

    Attributes:
        id: The dataset ID
        name: The dataset name
        description: The dataset description
        version_id: The current version ID
        examples: List of examples in this version
        metadata: Additional dataset metadata
        created_at: When the dataset was created
        updated_at: When the dataset was last updated
        example_count: Number of examples in this version
    """

    def __init__(
        self,
        dataset_info: Union[v1.Dataset, v1.DatasetWithExampleCount],
        examples_data: v1.ListDatasetExamplesData,
    ):
        self._dataset_info = dataset_info
        self._examples_data = examples_data

    @property
    def id(self) -> str:
        """The dataset ID."""
        return self._dataset_info["id"]

    @property
    def name(self) -> str:
        """The dataset name."""
        return self._dataset_info["name"]

    @property
    def description(self) -> Optional[str]:
        """The dataset description."""
        return self._dataset_info.get("description")

    @property
    def version_id(self) -> str:
        """The current version ID."""
        return self._examples_data["version_id"]

    @property
    def examples(self) -> list[v1.DatasetExample]:
        """List of examples in this version."""
        return list(self._examples_data["examples"])

    @property
    def metadata(self) -> dict[str, Any]:
        """Additional dataset metadata."""
        return dict(self._dataset_info.get("metadata", {}))

    @property
    def created_at(self) -> Optional[datetime]:
        """When the dataset was created."""
        if created_at := self._dataset_info.get("created_at"):
            return _parse_datetime(created_at)
        return None

    @property
    def updated_at(self) -> Optional[datetime]:
        """When the dataset was last updated."""
        if updated_at := self._dataset_info.get("updated_at"):
            return _parse_datetime(updated_at)
        return None

    @property
    def example_count(self) -> int:
        """Number of examples in this version."""
        if "example_count" in self._dataset_info:
            return self._dataset_info["example_count"]  # type: ignore[no-any-return,typeddict-item]
        return len(self.examples)

    def __repr__(self) -> str:
        return (
            f"Dataset(id={self.id!r}, name={self.name!r}, "
            f"version_id={self.version_id!r}, examples={self.example_count})"
        )

    def __len__(self) -> int:
        """Number of examples in this dataset version."""
        return len(self.examples)

    def __iter__(self) -> Iterator[v1.DatasetExample]:
        """Iterate over examples."""
        return iter(self.examples)

    def __getitem__(self, index: int) -> v1.DatasetExample:
        """Get example by index."""
        return self.examples[index]

    def to_dataframe(self) -> "pd.DataFrame":
        """
        Convert the dataset examples to a pandas DataFrame.

        Returns:
            A pandas DataFrame with the following columns:
            - input: Dictionary containing the input data
            - output: Dictionary containing the output data
            - metadata: Dictionary containing the metadata

            The DataFrame is indexed by example_id.

        Raises:
            ImportError: If pandas is not installed.

        Example:
            >>> dataset = client.datasets.get_dataset(dataset="my-dataset")
            >>> df = dataset.to_dataframe()
            >>> print(df.columns)
            Index(['input', 'output', 'metadata'], dtype='object')
            >>> print(df.index.name)
            example_id
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                "pandas is required to use to_dataframe(). " "Install it with 'pip install pandas'"
            )

        if not self.examples:
            return pd.DataFrame(columns=["input", "output", "metadata"]).set_index(  # pyright: ignore[reportUnknownMemberType]
                pd.Index([], name="example_id")
            )

        records = [
            {
                "example_id": example["id"],
                "input": deepcopy(example["input"]),
                "output": deepcopy(example["output"]),
                "metadata": deepcopy(example["metadata"]),
            }
            for example in self.examples
        ]

        return pd.DataFrame.from_records(records).set_index("example_id")  # pyright: ignore[reportUnknownMemberType]

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the dataset to a JSON-serializable dictionary.

        Example:
            >>> dataset = client.datasets.get_dataset(dataset="my-dataset")
            >>> json_data = dataset.to_dict()
            >>> restored = Dataset.from_dict(json_data)
        """
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "metadata": deepcopy(self.metadata),
            "created_at": self.created_at.isoformat() if self.created_at else None,
            "updated_at": self.updated_at.isoformat() if self.updated_at else None,
            # Additional fields from DatasetWithExampleCount and ListDatasetExamplesData
            "example_count": self.example_count,
            "version_id": self.version_id,
            "examples": deepcopy(self.examples),
        }

    @classmethod
    def from_dict(cls, json_data: dict[str, Any]) -> "Dataset":
        """
        Create a Dataset instance from a JSON-serializable dictionary.

        Args:
            json_data: Dictionary containing dataset information, typically from to_dict()

        Returns:
            A Dataset instance with the provided data

        Raises:
            ValueError: If required fields are missing from json_data
            KeyError: If json_data is missing required keys

        Example:
            >>> json_data = dataset.to_dict()
            >>> restored = Dataset.from_dict(json_data)
            >>> assert restored.id == dataset.id
        """
        required_fields = {"id", "name", "version_id", "examples"}
        if not all(field in json_data for field in required_fields):
            missing = required_fields - set(json_data.keys())
            raise ValueError(f"Missing required fields in json_data: {missing}")

        dataset_info = {
            "id": json_data["id"],
            "name": json_data["name"],
        }

        if json_data.get("description") is not None:
            dataset_info["description"] = json_data["description"]

        if json_data.get("metadata"):
            dataset_info["metadata"] = deepcopy(json_data["metadata"])

        if json_data.get("created_at"):
            dataset_info["created_at"] = json_data["created_at"]

        if json_data.get("updated_at"):
            dataset_info["updated_at"] = json_data["updated_at"]

        if json_data.get("example_count") is not None:
            dataset_info["example_count"] = json_data["example_count"]

        examples_data = {
            "version_id": json_data["version_id"],
            "examples": deepcopy(json_data["examples"]),
        }

        return cls(dataset_info, examples_data)  # type: ignore[arg-type]


# Type alias for flexible dataset identification
DatasetIdentifier = Union[
    str,  # dataset_id or dataset_name
    Dataset,  # Rich dataset object
    dict[str, Any],  # Dataset info dict from API
]


class DatasetKeys:
    """
    Validates dataset key specifications.
    """

    def __init__(
        self, input_keys: frozenset[str], output_keys: frozenset[str], metadata_keys: frozenset[str]
    ):
        self.input = input_keys
        self.output = output_keys
        self.metadata = metadata_keys

        if self.input & self.output:
            raise ValueError(f"Input and output keys overlap: {self.input & self.output}")
        if self.input & self.metadata:
            raise ValueError(f"Input and metadata keys overlap: {self.input & self.metadata}")
        if self.output & self.metadata:
            raise ValueError(f"Output and metadata keys overlap: {self.output & self.metadata}")

    def check_differences(self, available_keys: frozenset[str]) -> None:
        """Check that all specified keys exist in available keys."""
        all_keys = self.input | self.output | self.metadata
        if diff := all_keys - available_keys:
            raise ValueError(f"Keys not found in available columns: {diff}")

    def __iter__(self) -> "Iterator[str]":
        """Allow iteration over all keys."""
        return iter(self.input | self.output | self.metadata)


def _parse_datetime(datetime_str: str) -> datetime:
    """Convert ISO datetime string to datetime object."""
    return datetime.fromisoformat(datetime_str)


class Datasets:
    """
    Provides methods for interacting with dataset resources.

    Example:
        Basic usage:
            >>> from phoenix.client import Client
            >>> client = Client()
            >>> dataset = client.datasets.get_dataset(dataset="my-dataset")
            >>> print(f"Dataset {dataset.name} has {len(dataset)} examples")

        Creating and updating datasets:
            >>> # Create a new dataset
            >>> dataset = client.datasets.create_dataset(
            ...     name="qa-dataset",
            ...     inputs=[
            ...         {"question": "What is 2+2?"},
            ...         {"question": "What's the capital of France?"},
            ...     ],
            ...     outputs=[{"answer": "4"}, {"answer": "Paris"}]
            ... )
            >>>
            >>> # Add more examples later
            >>> updated = client.datasets.add_examples_to_dataset(
            ...     dataset="qa-dataset",
            ...     inputs=[{"question": "Who wrote Hamlet?"}],
            ...     outputs=[{"answer": "Shakespeare"}]
            ... )

        Working with DataFrames:
            >>> # Convert dataset to DataFrame
            >>> df = dataset.to_dataframe()
            >>> print(df.columns)  # Index(['input', 'output', 'metadata'], dtype='object')
            >>>
            >>> # Create dataset from DataFrame
            >>> import pandas as pd
            >>> df = pd.DataFrame({
            ...     "prompt": ["Hello", "Hi there"],
            ...     "response": ["Hi!", "Hello!"],
            ...     "score": [0.9, 0.95]
            ... })
            >>> dataset = client.datasets.create_dataset(
            ...     name="greetings",
            ...     dataframe=df,
            ...     input_keys=["prompt"],
            ...     output_keys=["response"],
            ...     metadata_keys=["score"]
            ... )
    """

    def __init__(self, client: httpx.Client) -> None:
        self._client = client

    def _resolve_dataset_id_and_name(
        self,
        dataset: DatasetIdentifier,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> tuple[Optional[str], Optional[str]]:
        """
        Resolve dataset ID and name from various input forms.

        Returns:
            Tuple of (dataset_id, dataset_name), where either or both may be None
        """
        if isinstance(dataset, Dataset):
            return dataset.id, dataset.name
        elif isinstance(dataset, str):
            if is_node_id(dataset, "Dataset"):
                return dataset, None
            else:
                return None, dataset
        elif isinstance(dataset, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
            return dataset.get("id"), dataset.get("name")
        else:
            raise ValueError(
                "Dataset must be a dataset ID string, name string, Dataset object, or dict"
            )

    def get_dataset(
        self,
        *,
        dataset: DatasetIdentifier,
        version_id: Optional[str] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Gets the dataset for a specific version, or gets the latest version of
        the dataset if no version is specified.

        Args:
            dataset: A dataset identifier - can be a dataset ID string, name string,
                Dataset object, or dict with 'id'/'name' fields.
            version_id: An ID for the version of the dataset, or None.
            timeout: Optional request timeout in seconds.

        Returns:
            A Dataset object containing the dataset metadata and examples.

        Raises:
            ValueError: If dataset format is invalid.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        resolved_id, resolved_name = self._resolve_dataset_id_and_name(dataset, timeout=timeout)

        if resolved_id:
            dataset_id = resolved_id
        elif resolved_name:
            dataset_id = self._get_dataset_id_by_name(dataset_name=resolved_name, timeout=timeout)
        else:
            # This shouldn't happen with current resolution logic, but just in case
            raise ValueError("Could not determine dataset ID or name from input")

        dataset_response = self._client.get(
            url=f"v1/datasets/{quote(dataset_id)}",
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        dataset_response.raise_for_status()
        dataset_info = dataset_response.json()["data"]

        params = {"version_id": version_id} if version_id else None
        examples_response = self._client.get(
            url=f"v1/datasets/{quote(dataset_id)}/examples",
            params=params,
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        examples_response.raise_for_status()
        examples_data = examples_response.json()["data"]

        return Dataset(dataset_info, examples_data)

    def get_dataset_versions(
        self,
        *,
        dataset: DatasetIdentifier,
        limit: Optional[int] = 100,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> list[v1.DatasetVersion]:
        """
        Get dataset versions as a list of dictionaries.

        Args:
            dataset: A dataset identifier - can be a dataset ID string, name string,
                Dataset object, or dict with 'id'/'name' fields.
            limit: Maximum number of versions to return, starting from the most recent version
            timeout: Optional request timeout in seconds.

        Returns:
            List of dictionaries containing version information, including:
                - version_id: The version ID
                - created_at: When the version was created (as datetime object)
                - description: Version description (if any)
                - metadata: Version metadata (if any)

        Raises:
            ValueError: If dataset format is invalid.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        resolved_id, resolved_name = self._resolve_dataset_id_and_name(dataset, timeout=timeout)

        if resolved_id:
            dataset_id = resolved_id
        elif resolved_name:
            dataset_id = self._get_dataset_id_by_name(dataset_name=resolved_name, timeout=timeout)
        else:
            raise ValueError("Could not determine dataset ID or name from input")

        response = self._client.get(
            url=f"v1/datasets/{dataset_id}/versions",
            params={"limit": limit},
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        response.raise_for_status()

        records = response.json()["data"]
        if not records:
            return []

        for record in records:
            if "created_at" in record:
                record["created_at"] = _parse_datetime(record["created_at"])

        return records  # type: ignore[no-any-return]

    def create_dataset(
        self,
        *,
        name: str,
        examples: Optional[Union[v1.DatasetExample, Iterable[v1.DatasetExample]]] = None,
        dataframe: Optional["pd.DataFrame"] = None,
        csv_file_path: Optional[Union[str, Path]] = None,
        input_keys: Iterable[str] = (),
        output_keys: Iterable[str] = (),
        metadata_keys: Iterable[str] = (),
        inputs: Iterable[Mapping[str, Any]] = (),
        outputs: Iterable[Mapping[str, Any]] = (),
        metadata: Iterable[Mapping[str, Any]] = (),
        dataset_description: Optional[str] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Create a new dataset by uploading examples to the Phoenix server.

        Args:
            dataset_name: Name of the dataset.
            examples: Either a single DatasetExample or list of DatasetExample objects to add.
                When provided, inputs/outputs/metadata are extracted automatically.
            dataframe: pandas DataFrame (requires pandas to be installed).
            csv_file_path: Location of a CSV text file
            input_keys: List of column names used as input keys.
            output_keys: List of column names used as output keys.
            metadata_keys: List of column names used as metadata keys.
            inputs: List of dictionaries each corresponding to an example.
            outputs: List of dictionaries each corresponding to an example.
            metadata: List of dictionaries each corresponding to an example.
            dataset_description: Description of the dataset.
            timeout: Optional request timeout in seconds.

        Returns:
            A Dataset object containing the uploaded dataset and examples.

        Raises:
            ValueError: If invalid parameter combinations are provided.
            ImportError: If pandas is required but not installed.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        has_examples = examples is not None
        has_tabular = dataframe is not None or csv_file_path is not None
        has_json = any(inputs) or any(outputs) or any(metadata)

        if sum([has_examples, has_tabular, has_json]) > 1:
            raise ValueError(
                "Please provide only one of: examples, tabular data (dataframe/csv_file_path), "
                "or dictionaries (inputs/outputs/metadata)"
            )

        if dataframe is not None and csv_file_path is not None:
            raise ValueError("Please provide either dataframe or csv_file_path, but not both")

        if examples is not None:
            examples_list: list[v1.DatasetExample]
            if _is_valid_dataset_example(examples):
                examples_list = [examples]  # type: ignore[list-item]
            else:
                examples_list = list(examples)  # type: ignore[arg-type]

            inputs = [dict(example["input"]) for example in examples_list]
            outputs = [dict(example["output"]) for example in examples_list]
            metadata = [dict(example["metadata"]) for example in examples_list]

        if has_tabular:
            table = dataframe if dataframe is not None else csv_file_path
            assert table is not None
            return self._upload_tabular_dataset(
                table,
                dataset_name=name,
                input_keys=input_keys,
                output_keys=output_keys,
                metadata_keys=metadata_keys,
                dataset_description=dataset_description,
                action="create",
                timeout=timeout,
            )
        else:
            return self._upload_json_dataset(
                dataset_name=name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                dataset_description=dataset_description,
                action="create",
                timeout=timeout,
            )

    def add_examples_to_dataset(
        self,
        *,
        dataset: DatasetIdentifier,
        examples: Optional[Union[v1.DatasetExample, Iterable[v1.DatasetExample]]] = None,
        dataframe: Optional["pd.DataFrame"] = None,
        csv_file_path: Optional[Union[str, Path]] = None,
        input_keys: Iterable[str] = (),
        output_keys: Iterable[str] = (),
        metadata_keys: Iterable[str] = (),
        inputs: Iterable[Mapping[str, Any]] = (),
        outputs: Iterable[Mapping[str, Any]] = (),
        metadata: Iterable[Mapping[str, Any]] = (),
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Append examples to an existing dataset on the Phoenix server.

        Args:
            dataset: A dataset identifier - can be a dataset ID string, name string,
                Dataset object, or dict with 'id'/'name' fields.
            examples: Either a single DatasetExample or list of DatasetExample objects to add.
                When provided, inputs/outputs/metadata are extracted automatically.
            dataframe: pandas DataFrame (requires pandas to be installed).
            csv_file_path: Location of a CSV text file
            input_keys: List of column names used as input keys.
            output_keys: List of column names used as output keys.
            metadata_keys: List of column names used as metadata keys.
            inputs: List of dictionaries each corresponding to an example.
            outputs: List of dictionaries each corresponding to an example.
            metadata: List of dictionaries each corresponding to an example.
            timeout: Optional request timeout in seconds.

        Returns:
            A Dataset object containing the updated dataset and examples.

        Raises:
            ValueError: If invalid parameter combinations are provided.
            ImportError: If pandas is required but not installed.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        resolved_id, resolved_name = self._resolve_dataset_id_and_name(dataset, timeout=timeout)

        if not resolved_name:
            if resolved_id:
                response = self._client.get(
                    url=f"v1/datasets/{quote(resolved_id)}",
                    headers={"accept": "application/json"},
                    timeout=timeout,
                )
                response.raise_for_status()
                resolved_name = response.json()["data"]["name"]
            else:
                raise ValueError("Could not determine dataset name from input")

        # At this point resolved_name is guaranteed to be not None
        assert resolved_name is not None

        has_examples = examples is not None
        has_tabular = dataframe is not None or csv_file_path is not None
        has_json = any(inputs) or any(outputs) or any(metadata)

        if sum([has_examples, has_tabular, has_json]) > 1:
            raise ValueError(
                "Please provide only one of: examples, tabular data (dataframe/csv_file_path), "
                "or dictionaries (inputs/outputs/metadata)"
            )

        if dataframe is not None and csv_file_path is not None:
            raise ValueError("Please provide either dataframe or csv_file_path, but not both")

        if examples is not None:
            examples_list: list[v1.DatasetExample]
            if _is_valid_dataset_example(examples):
                examples_list = [examples]  # type: ignore[list-item]
            else:
                examples_list = list(examples)  # type: ignore[arg-type]

            inputs = [dict(example["input"]) for example in examples_list]
            outputs = [dict(example["output"]) for example in examples_list]
            metadata = [dict(example["metadata"]) for example in examples_list]

        if has_tabular:
            table = dataframe if dataframe is not None else csv_file_path
            assert table is not None
            return self._upload_tabular_dataset(
                table,
                dataset_name=resolved_name,
                input_keys=input_keys,
                output_keys=output_keys,
                metadata_keys=metadata_keys,
                dataset_description=None,
                action="append",
                timeout=timeout,
            )
        else:
            return self._upload_json_dataset(
                dataset_name=resolved_name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                dataset_description=None,
                action="append",
                timeout=timeout,
            )

    def _get_dataset_id_by_name(
        self,
        *,
        dataset_name: str,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> str:
        """
        Gets a dataset ID by name.

        Args:
            dataset_name: The name of the dataset.
            timeout: Optional request timeout in seconds.

        Returns:
            The dataset ID.

        Raises:
            ValueError: If dataset not found or multiple datasets found.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        response = self._client.get(
            url="v1/datasets",
            params={"name": dataset_name},
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        response.raise_for_status()

        records = response.json()["data"]
        if not records:
            raise ValueError(f"Dataset not found: {dataset_name}")
        if len(records) > 1:
            raise ValueError(f"Multiple datasets found with name: {dataset_name}")

        return str(records[0]["id"])

    def _upload_tabular_dataset(
        self,
        table: Union[str, Path, "pd.DataFrame"],
        *,
        dataset_name: str,
        input_keys: Iterable[str],
        output_keys: Iterable[str] = (),
        metadata_keys: Iterable[str] = (),
        dataset_description: Optional[str] = None,
        action: Literal["create", "append"] = "create",
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Upload tabular data (CSV or DataFrame) as dataset.
        """
        input_keys_set = frozenset(input_keys)
        output_keys_set = frozenset(output_keys)
        metadata_keys_set = frozenset(metadata_keys)

        # Auto-infer keys if none provided
        if not any([input_keys_set, output_keys_set, metadata_keys_set]):
            input_keys_tuple, output_keys_tuple, metadata_keys_tuple = _infer_keys(table)
            input_keys_set = frozenset(input_keys_tuple)
            output_keys_set = frozenset(output_keys_tuple)
            metadata_keys_set = frozenset(metadata_keys_tuple)

        keys = DatasetKeys(input_keys_set, output_keys_set, metadata_keys_set)

        if isinstance(table, Path) or isinstance(table, str):
            file = _prepare_csv(Path(table), keys)
        else:
            try:
                import pandas as pd

                if not isinstance(table, pd.DataFrame):  # pyright: ignore[reportUnnecessaryIsInstance]
                    raise ValueError("Expected pandas DataFrame")
            except ImportError:
                raise ImportError(
                    "pandas is required to upload DataFrames. "
                    "Install it with 'pip install pandas'"
                )
            file = _prepare_dataframe_as_csv(table, keys)

        logger.info("Uploading dataset...")
        response = self._client.post(
            url="v1/datasets/upload",
            files={"file": file},
            data={
                "action": action,
                "name": dataset_name,
                "description": dataset_description or "",
                "input_keys[]": sorted(keys.input),
                "output_keys[]": sorted(keys.output),
                "metadata_keys[]": sorted(keys.metadata),
            },
            params={"sync": True},
            headers={"accept": "application/json"},
            timeout=timeout,
        )

        return self._process_dataset_upload_response(response, timeout=timeout)

    def _upload_json_dataset(
        self,
        *,
        dataset_name: str,
        inputs: Iterable[Mapping[str, Any]],
        outputs: Iterable[Mapping[str, Any]] = (),
        metadata: Iterable[Mapping[str, Any]] = (),
        dataset_description: Optional[str] = None,
        action: Literal["create", "append"] = "create",
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Upload JSON data as dataset.
        """
        # Convert to lists to handle generators and validate
        inputs_list = list(inputs)
        outputs_list = list(outputs) if outputs else []
        metadata_list = list(metadata) if metadata else []

        if not inputs_list:
            raise ValueError("inputs must be non-empty")

        if not _is_all_dict(inputs_list):
            raise ValueError("inputs must contain only dictionaries")

        for name, data in [("outputs", outputs_list), ("metadata", metadata_list)]:
            if data:
                if len(data) != len(inputs_list):
                    raise ValueError(
                        f"{name} must have same length as inputs "
                        f"({len(data)} != {len(inputs_list)})"
                    )
                if not _is_all_dict(data):
                    raise ValueError(f"{name} must contain only dictionaries")

        payload = {
            "action": action,
            "name": dataset_name,
            "inputs": inputs_list,
            "outputs": outputs_list or [{}] * len(inputs_list),
            "metadata": metadata_list or [{}] * len(inputs_list),
        }

        if dataset_description is not None:
            payload["description"] = dataset_description

        logger.info("Uploading dataset...")
        response = self._client.post(
            url="v1/datasets/upload",
            json=payload,
            params={"sync": True},
            headers={"accept": "application/json"},
            timeout=timeout,
        )

        return self._process_dataset_upload_response(response, timeout=timeout)

    def _process_dataset_upload_response(
        self,
        response: httpx.Response,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Process the response from dataset upload operations.
        """
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            try:
                error_detail = response.json().get("detail", str(e))
            except Exception:
                error_detail = response.text or str(e)
            raise DatasetUploadError(f"Dataset upload failed: {error_detail}") from e

        upload_data = response.json()["data"]
        dataset_id = upload_data["dataset_id"]
        version_id = upload_data["version_id"]

        dataset = self.get_dataset(
            dataset=dataset_id,
            version_id=version_id,
            timeout=timeout,
        )

        logger.info(f"Dataset uploaded successfully. ID: {dataset_id}, Version: {version_id}")

        return dataset


class AsyncDatasets:
    """
    Provides async methods for interacting with dataset resources.

    Example:
        Basic usage:
            >>> from phoenix.client import AsyncClient
            >>> client = AsyncClient()
            >>> dataset = await client.datasets.get_dataset(dataset="my-dataset")
            >>> print(f"Dataset {dataset.name} has {len(dataset)} examples")

        Creating and updating datasets:
            >>> # Create a new dataset
            >>> dataset = await client.datasets.create_dataset(
            ...     name="qa-dataset",
            ...     inputs=[
            ...         {"question": "What is 2+2?"},
            ...         {"question": "What's the capital of France?"},
            ...     ],
            ...     outputs=[{"answer": "4"}, {"answer": "Paris"}]
            ... )
            >>>
            >>> # Add more examples later
            >>> updated = await client.datasets.add_examples_to_dataset(
            ...     dataset="qa-dataset",
            ...     inputs=[{"question": "Who wrote Hamlet?"}],
            ...     outputs=[{"answer": "Shakespeare"}]
            ... )

        Working with DataFrames:
            >>> # Convert dataset to DataFrame (sync operation)
            >>> df = dataset.to_dataframe()
            >>> print(df.columns)  # Index(['input', 'output', 'metadata'], dtype='object')
            >>>
            >>> # Create dataset from DataFrame
            >>> import pandas as pd
            >>> df = pd.DataFrame({
            ...     "prompt": ["Hello", "Hi there"],
            ...     "response": ["Hi!", "Hello!"],
            ...     "score": [0.9, 0.95]
            ... })
            >>> dataset = await client.datasets.create_dataset(
            ...     name="greetings",
            ...     dataframe=df,
            ...     input_keys=["prompt"],
            ...     output_keys=["response"],
            ...     metadata_keys=["score"]
            ... )
    """

    def __init__(self, client: httpx.AsyncClient) -> None:
        self._client = client

    async def _resolve_dataset_id_and_name(
        self,
        dataset: DatasetIdentifier,
    ) -> tuple[Optional[str], Optional[str]]:
        """
        Resolve dataset ID and name from various input forms.

        Returns:
            Tuple of (dataset_id, dataset_name), where either or both may be None
        """
        if isinstance(dataset, Dataset):
            return dataset.id, dataset.name
        elif isinstance(dataset, str):
            if is_node_id(dataset, "Dataset"):
                return dataset, None
            else:
                return None, dataset
        elif isinstance(dataset, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
            return dataset.get("id"), dataset.get("name")
        else:
            raise ValueError(
                "Dataset must be a dataset ID string, name string, Dataset object, or dict"
            )

    async def get_dataset(
        self,
        *,
        dataset: DatasetIdentifier,
        version_id: Optional[str] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Gets the dataset for a specific version, or gets the latest version of
        the dataset if no version is specified.

        Args:
            dataset: A dataset identifier - can be a dataset ID string, name string,
                Dataset object, or dict with 'id'/'name' fields.
            version_id: An ID for the version of the dataset, or None.
            timeout: Optional request timeout in seconds.

        Returns:
            A Dataset object containing the dataset metadata and examples.

        Raises:
            ValueError: If dataset format is invalid.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        resolved_id, resolved_name = await self._resolve_dataset_id_and_name(dataset)

        if resolved_id:
            dataset_id = resolved_id
        elif resolved_name:
            dataset_id = await self._get_dataset_id_by_name(
                dataset_name=resolved_name, timeout=timeout
            )
        else:
            # This shouldn't happen with current resolution logic, but just in case
            raise ValueError("Could not determine dataset ID or name from input")

        dataset_response = await self._client.get(
            url=f"v1/datasets/{quote(dataset_id)}",
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        dataset_response.raise_for_status()
        dataset_info = dataset_response.json()["data"]

        params = {"version_id": version_id} if version_id else None
        examples_response = await self._client.get(
            url=f"v1/datasets/{quote(dataset_id)}/examples",
            params=params,
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        examples_response.raise_for_status()
        examples_data = examples_response.json()["data"]

        return Dataset(dataset_info, examples_data)

    async def get_dataset_versions(
        self,
        *,
        dataset: DatasetIdentifier,
        limit: Optional[int] = 100,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> list[v1.DatasetVersion]:
        """
        Get dataset versions as a list of dictionaries.

        Args:
            dataset: A dataset identifier - can be a dataset ID string, name string,
                Dataset object, or dict with 'id'/'name' fields.
            limit: Maximum number of versions to return, starting from the most recent version
            timeout: Optional request timeout in seconds.

        Returns:
            List of dictionaries containing version information, including:
                - version_id: The version ID
                - created_at: When the version was created (as datetime object)
                - description: Version description (if any)
                - metadata: Version metadata (if any)

        Raises:
            ValueError: If dataset format is invalid.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        resolved_id, resolved_name = await self._resolve_dataset_id_and_name(dataset)

        if resolved_id:
            dataset_id = resolved_id
        elif resolved_name:
            dataset_id = await self._get_dataset_id_by_name(
                dataset_name=resolved_name, timeout=timeout
            )
        else:
            raise ValueError("Could not determine dataset ID or name from input")

        response = await self._client.get(
            url=f"v1/datasets/{dataset_id}/versions",
            params={"limit": limit},
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        response.raise_for_status()

        records = response.json()["data"]
        if not records:
            return []

        for record in records:
            if "created_at" in record:
                record["created_at"] = _parse_datetime(record["created_at"])

        return records  # type: ignore[no-any-return]

    async def create_dataset(
        self,
        *,
        name: str,
        examples: Optional[Union[v1.DatasetExample, Iterable[v1.DatasetExample]]] = None,
        dataframe: Optional["pd.DataFrame"] = None,
        csv_file_path: Optional[Union[str, Path]] = None,
        input_keys: Iterable[str] = (),
        output_keys: Iterable[str] = (),
        metadata_keys: Iterable[str] = (),
        inputs: Iterable[Mapping[str, Any]] = (),
        outputs: Iterable[Mapping[str, Any]] = (),
        metadata: Iterable[Mapping[str, Any]] = (),
        dataset_description: Optional[str] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Create a new dataset by uploading examples to the Phoenix server.

        Args:
            dataset_name: Name of the dataset.
            examples: Either a single DatasetExample or list of DatasetExample objects to add.
                When provided, inputs/outputs/metadata are extracted automatically.
            dataframe: pandas DataFrame (requires pandas to be installed).
            csv_file_path: Location of a CSV text file
            input_keys: List of column names used as input keys.
            output_keys: List of column names used as output keys.
            metadata_keys: List of column names used as metadata keys.
            inputs: List of dictionaries each corresponding to an example.
            outputs: List of dictionaries each corresponding to an example.
            metadata: List of dictionaries each corresponding to an example.
            dataset_description: Description of the dataset.
            timeout: Optional request timeout in seconds.

        Returns:
            A Dataset object containing the uploaded dataset and examples.

        Raises:
            ValueError: If invalid parameter combinations are provided.
            ImportError: If pandas is required but not installed.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        has_examples = examples is not None
        has_tabular = dataframe is not None or csv_file_path is not None
        has_json = any(inputs) or any(outputs) or any(metadata)

        if sum([has_examples, has_tabular, has_json]) > 1:
            raise ValueError(
                "Please provide only one of: examples, tabular data (dataframe/csv_file_path), "
                "or JSON data (inputs/outputs/metadata)"
            )

        if dataframe is not None and csv_file_path is not None:
            raise ValueError("Please provide either dataframe or csv_file_path, but not both")

        if examples is not None:
            examples_list: list[v1.DatasetExample]
            if _is_valid_dataset_example(examples):
                examples_list = [examples]  # type: ignore[list-item]
            else:
                examples_list = list(examples)  # type: ignore[arg-type]

            inputs = [dict(example["input"]) for example in examples_list]
            outputs = [dict(example["output"]) for example in examples_list]
            metadata = [dict(example["metadata"]) for example in examples_list]

        if has_tabular:
            table = dataframe if dataframe is not None else csv_file_path
            assert table is not None
            return await self._upload_tabular_dataset(
                table,
                dataset_name=name,
                input_keys=input_keys,
                output_keys=output_keys,
                metadata_keys=metadata_keys,
                dataset_description=dataset_description,
                action="create",
                timeout=timeout,
            )
        else:
            return await self._upload_json_dataset(
                dataset_name=name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                dataset_description=dataset_description,
                action="create",
                timeout=timeout,
            )

    async def add_examples_to_dataset(
        self,
        *,
        dataset: DatasetIdentifier,
        examples: Optional[Union[v1.DatasetExample, Iterable[v1.DatasetExample]]] = None,
        dataframe: Optional["pd.DataFrame"] = None,
        csv_file_path: Optional[Union[str, Path]] = None,
        input_keys: Iterable[str] = (),
        output_keys: Iterable[str] = (),
        metadata_keys: Iterable[str] = (),
        inputs: Iterable[Mapping[str, Any]] = (),
        outputs: Iterable[Mapping[str, Any]] = (),
        metadata: Iterable[Mapping[str, Any]] = (),
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """
        Append examples to an existing dataset on the Phoenix server.

        Args:
            dataset: A dataset identifier - can be a dataset ID string, name string,
                Dataset object, or dict with 'id'/'name' fields.
            examples: Either a single DatasetExample or list of DatasetExample objects to add.
                When provided, inputs/outputs/metadata are extracted automatically.
            dataframe: pandas DataFrame (requires pandas to be installed).
            csv_file_path: Location of a CSV text file
            input_keys: List of column names used as input keys.
            output_keys: List of column names used as output keys.
            metadata_keys: List of column names used as metadata keys.
            inputs: List of dictionaries each corresponding to an example.
            outputs: List of dictionaries each corresponding to an example.
            metadata: List of dictionaries each corresponding to an example.
            timeout: Optional request timeout in seconds.

        Returns:
            A Dataset object containing the updated dataset and examples.

        Raises:
            ValueError: If invalid parameter combinations are provided.
            ImportError: If pandas is required but not installed.
            httpx.HTTPStatusError: If the API returns an error response.
        """
        resolved_id, resolved_name = await self._resolve_dataset_id_and_name(dataset)

        if not resolved_name:
            if resolved_id:
                response = await self._client.get(
                    url=f"v1/datasets/{quote(resolved_id)}",
                    headers={"accept": "application/json"},
                    timeout=timeout,
                )
                response.raise_for_status()
                resolved_name = response.json()["data"]["name"]
            else:
                raise ValueError("Could not determine dataset name from input")

        # At this point resolved_name is guaranteed to be not None
        assert resolved_name is not None

        has_examples = examples is not None
        has_tabular = dataframe is not None or csv_file_path is not None
        has_json = any(inputs) or any(outputs) or any(metadata)

        if sum([has_examples, has_tabular, has_json]) > 1:
            raise ValueError(
                "Please provide only one of: examples, tabular data (dataframe/csv_file_path), "
                "or JSON data (inputs/outputs/metadata)"
            )

        if dataframe is not None and csv_file_path is not None:
            raise ValueError("Please provide either dataframe or csv_file_path, but not both")

        if examples is not None:
            examples_list: list[v1.DatasetExample]
            if _is_valid_dataset_example(examples):
                examples_list = [examples]  # type: ignore[list-item]
            else:
                examples_list = list(examples)  # type: ignore[arg-type]

            inputs = [dict(example["input"]) for example in examples_list]
            outputs = [dict(example["output"]) for example in examples_list]
            metadata = [dict(example["metadata"]) for example in examples_list]

        if has_tabular:
            table = dataframe if dataframe is not None else csv_file_path
            assert table is not None
            return await self._upload_tabular_dataset(
                table,
                dataset_name=resolved_name,
                input_keys=input_keys,
                output_keys=output_keys,
                metadata_keys=metadata_keys,
                dataset_description=None,
                action="append",
                timeout=timeout,
            )
        else:
            return await self._upload_json_dataset(
                dataset_name=resolved_name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                dataset_description=None,
                action="append",
                timeout=timeout,
            )

    async def _get_dataset_id_by_name(
        self,
        *,
        dataset_name: str,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> str:
        """Async version of _get_dataset_id_by_name."""
        response = await self._client.get(
            url="v1/datasets",
            params={"name": dataset_name},
            headers={"accept": "application/json"},
            timeout=timeout,
        )
        response.raise_for_status()

        records = response.json()["data"]
        if not records:
            raise ValueError(f"Dataset not found: {dataset_name}")
        if len(records) > 1:
            raise ValueError(f"Multiple datasets found with name: {dataset_name}")

        return str(records[0]["id"])

    async def _upload_tabular_dataset(
        self,
        table: Union[str, Path, "pd.DataFrame"],
        *,
        dataset_name: str,
        input_keys: Iterable[str],
        output_keys: Iterable[str] = (),
        metadata_keys: Iterable[str] = (),
        dataset_description: Optional[str] = None,
        action: Literal["create", "append"] = "create",
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """Async version of _upload_tabular_dataset."""
        input_keys_set = frozenset(input_keys)
        output_keys_set = frozenset(output_keys)
        metadata_keys_set = frozenset(metadata_keys)

        # Auto-infer keys if none provided
        if not any([input_keys_set, output_keys_set, metadata_keys_set]):
            input_keys_tuple, output_keys_tuple, metadata_keys_tuple = _infer_keys(table)
            input_keys_set = frozenset(input_keys_tuple)
            output_keys_set = frozenset(output_keys_tuple)
            metadata_keys_set = frozenset(metadata_keys_tuple)

        keys = DatasetKeys(input_keys_set, output_keys_set, metadata_keys_set)

        if isinstance(table, Path) or isinstance(table, str):
            file = _prepare_csv(Path(table), keys)
        else:
            try:
                import pandas as pd

                if not isinstance(table, pd.DataFrame):  # pyright: ignore[reportUnnecessaryIsInstance]
                    raise ValueError("Expected pandas DataFrame")
            except ImportError:
                raise ImportError(
                    "pandas is required to upload DataFrames. "
                    "Install it with 'pip install pandas'"
                )
            file = _prepare_dataframe_as_csv(table, keys)

        logger.info("Uploading dataset...")
        response = await self._client.post(
            url="v1/datasets/upload",
            files={"file": file},
            data={
                "action": action,
                "name": dataset_name,
                "description": dataset_description or "",
                "input_keys[]": sorted(keys.input),
                "output_keys[]": sorted(keys.output),
                "metadata_keys[]": sorted(keys.metadata),
            },
            params={"sync": True},
            headers={"accept": "application/json"},
            timeout=timeout,
        )

        return await self._process_dataset_upload_response(response, timeout=timeout)

    async def _upload_json_dataset(
        self,
        *,
        dataset_name: str,
        inputs: Iterable[Mapping[str, Any]],
        outputs: Iterable[Mapping[str, Any]] = (),
        metadata: Iterable[Mapping[str, Any]] = (),
        dataset_description: Optional[str] = None,
        action: Literal["create", "append"] = "create",
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """Async version of _upload_json_dataset."""
        # Convert to lists to handle generators and validate
        inputs_list = list(inputs)
        outputs_list = list(outputs) if outputs else []
        metadata_list = list(metadata) if metadata else []

        if not inputs_list:
            raise ValueError("inputs must be non-empty")

        if not _is_all_dict(inputs_list):
            raise ValueError("inputs must contain only dictionaries")

        for name, data in [("outputs", outputs_list), ("metadata", metadata_list)]:
            if data:
                if len(data) != len(inputs_list):
                    raise ValueError(
                        f"{name} must have same length as inputs "
                        f"({len(data)} != {len(inputs_list)})"
                    )
                if not _is_all_dict(data):
                    raise ValueError(f"{name} must contain only dictionaries")

        payload = {
            "action": action,
            "name": dataset_name,
            "inputs": inputs_list,
            "outputs": outputs_list or [{}] * len(inputs_list),
            "metadata": metadata_list or [{}] * len(inputs_list),
        }

        if dataset_description is not None:
            payload["description"] = dataset_description

        logger.info("Uploading dataset...")
        response = await self._client.post(
            url="v1/datasets/upload",
            json=payload,
            params={"sync": True},
            headers={"accept": "application/json"},
            timeout=timeout,
        )

        return await self._process_dataset_upload_response(response, timeout=timeout)

    async def _process_dataset_upload_response(
        self,
        response: httpx.Response,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT_IN_SECONDS,
    ) -> Dataset:
        """Async version of _process_dataset_upload_response."""
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            try:
                error_detail = response.json().get("detail", str(e))
            except Exception:
                error_detail = response.text or str(e)
            raise DatasetUploadError(f"Dataset upload failed: {error_detail}") from e

        upload_data = response.json()["data"]
        dataset_id = upload_data["dataset_id"]
        version_id = upload_data["version_id"]

        dataset = await self.get_dataset(
            dataset=dataset_id,
            version_id=version_id,
            timeout=timeout,
        )

        logger.info(f"Dataset uploaded successfully. ID: {dataset_id}, Version: {version_id}")

        return dataset


def _get_csv_column_headers(path: Path) -> tuple[str, ...]:
    """
    Extract column headers from CSV file.
    """
    path = path.resolve()
    if not path.is_file():
        raise FileNotFoundError(f"File does not exist: {path}")

    with open(path, "r") as f:
        reader = csv.reader(f)
        try:
            column_headers = tuple(next(reader))
            # Check if there's at least one data row
            next(reader)
        except StopIteration as e:
            raise ValueError("CSV file has no data rows") from e

    return column_headers


def _prepare_csv(path: Path, keys: DatasetKeys) -> tuple[str, BinaryIO, str, dict[str, str]]:
    """
    Prepare CSV file for upload with validation and compression.
    """
    column_headers = _get_csv_column_headers(path)
    header_counts = Counter(column_headers)
    duplicates = [h for h, count in header_counts.items() if count > 1]
    if duplicates:
        raise ValueError(f"Duplicate column headers in CSV: {duplicates}")

    keys.check_differences(frozenset(column_headers))

    with open(path, "rb") as f:
        content = f.read()

    compressed = BytesIO()
    compressed.write(gzip.compress(content))
    compressed.seek(0)

    return (path.name, compressed, "text/csv", {"Content-Encoding": "gzip"})


def _prepare_dataframe_as_csv(
    df: "pd.DataFrame", keys: DatasetKeys
) -> tuple[str, BinaryIO, str, dict[str, str]]:
    """
    Prepare pandas DataFrame for upload as compressed CSV.
    """
    if df.empty:
        raise ValueError("DataFrame has no data")

    column_counts = Counter(df.columns)
    duplicates = [col for col, count in column_counts.items() if count > 1]
    if duplicates:
        raise ValueError(f"Duplicate column names in DataFrame: {duplicates}")

    keys.check_differences(frozenset(df.columns))

    # Ensure consistent column ordering: input, output, metadata
    selected_columns = sorted(keys.input) + sorted(keys.output) + sorted(keys.metadata)

    csv_buffer = BytesIO()
    df[selected_columns].to_csv(csv_buffer, index=False)
    csv_buffer.seek(0)

    compressed = BytesIO()
    compressed.write(gzip.compress(csv_buffer.read()))
    compressed.seek(0)

    return ("dataframe.csv", compressed, "text/csv", {"Content-Encoding": "gzip"})


def _infer_keys(
    table: Union[str, Path, "pd.DataFrame"],
) -> tuple[tuple[str, ...], tuple[str, ...], tuple[str, ...]]:
    """
    Infer input/output/metadata keys from table structure.

    Uses pattern matching to detect response/output columns automatically.
    """
    try:
        import pandas as pd

        if isinstance(table, pd.DataFrame):
            column_headers = tuple(table.columns)
        else:
            column_headers = _get_csv_column_headers(Path(table))
    except ImportError as e:
        if not isinstance(table, (str, Path)):
            raise ValueError("Pandas not available, table must be a CSV file path") from e
        column_headers = _get_csv_column_headers(Path(table))

    # Pattern to match output/response columns
    output_pattern = re.compile(r"(?i)(response|answer|output)s?$")

    output_idx = None
    for i, header in enumerate(column_headers):
        if output_pattern.search(header):
            output_idx = i
            break

    if output_idx is None:
        # No output column found - all columns are inputs
        return (column_headers, (), ())

    # Split columns: inputs before output, output column, metadata after
    input_cols = column_headers[:output_idx]
    output_cols = (column_headers[output_idx],)
    metadata_cols = column_headers[output_idx + 1 :]

    return (input_cols, output_cols, metadata_cols)


def _is_all_dict(seq: Iterable[Any]) -> bool:
    """Check if all items in sequence are dictionaries."""
    return all(isinstance(item, dict) for item in seq)


class DatasetUploadError(Exception):
    """Custom exception for dataset upload errors."""

    pass



from __future__ import annotations

from abc import ABC
from collections import abc
from dataclasses import dataclass
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Union,
)

from typing_extensions import Self, TypeAlias, assert_never

from phoenix.client.__generated__ import v1
from phoenix.client.helpers.sdk.anthropic.messages import (
    AnthropicMessageModelKwargs,
    create_prompt_version_from_anthropic,
)
from phoenix.client.helpers.sdk.anthropic.messages import (
    to_chat_messages_and_kwargs as to_messages_anthropic,
)
from phoenix.client.helpers.sdk.google_generativeai.generate_content import (
    GoogleModelKwargs,
    create_prompt_version_from_google,
)
from phoenix.client.helpers.sdk.google_generativeai.generate_content import (
    to_chat_messages_and_kwargs as to_messages_google,
)
from phoenix.client.helpers.sdk.openai.chat import (
    OpenAIChatCompletionModelKwargs,
    create_prompt_version_from_openai,
)
from phoenix.client.helpers.sdk.openai.chat import (
    to_chat_messages_and_kwargs as to_messages_openai,
)
from phoenix.client.utils.template_formatters import TemplateFormatter

if TYPE_CHECKING:
    from anthropic.types import MessageParam
    from anthropic.types.message_create_params import MessageCreateParamsBase
    from google.generativeai import protos
    from openai.types.chat import ChatCompletionMessageParam
    from openai.types.chat.completion_create_params import CompletionCreateParamsBase


class PromptVersion:
    """
    Represents a version of a prompt for different model providers.
    """

    def __init__(
        self,
        prompt: Sequence[v1.PromptMessage],
        /,
        *,
        model_name: str,
        description: Optional[str] = None,
        model_provider: Literal[
            "OPENAI", "AZURE_OPENAI", "ANTHROPIC", "GOOGLE", "DEEPSEEK", "XAI", "AWS", "OLLAMA"
        ] = "OPENAI",
        template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = "MUSTACHE",
    ) -> None:
        """
        Initializes a PromptVersion for syncing and template application

        Args:
            prompt (Sequence[v1.PromptMessage]): A sequence of prompt messages.
            model_name (str): The name of the model to use for the prompt.
            description (Optional[str]): A description of the prompt. Defaults
            to None. model_provider (Literal["OPENAI", "AZURE_OPENAI",
            "ANTHROPIC", "GOOGLE",
                "DEEPSEEK", "XAI", "AWS", "OLLAMA"]): The provider of the model
                to use for the prompt. Defaults to "OPENAI".
            template_format (Literal["F_STRING", "MUSTACHE", "NONE"]): The
            format of the template
                to use for the prompt. Defaults to "MUSTACHE".
        """
        self._template = v1.PromptChatTemplate(messages=prompt, type="chat")
        self._template_type: Literal["CHAT"] = "CHAT"
        self._model_name = model_name
        self._model_provider: Literal[
            "OPENAI", "AZURE_OPENAI", "ANTHROPIC", "GOOGLE", "DEEPSEEK", "XAI", "AWS", "OLLAMA"
        ] = model_provider
        self._template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = template_format
        self._description = description
        self._invocation_parameters: Union[
            v1.PromptOpenAIInvocationParameters,
            v1.PromptAzureOpenAIInvocationParameters,
            v1.PromptAnthropicInvocationParameters,
            v1.PromptGoogleInvocationParameters,
            v1.PromptDeepSeekInvocationParameters,
            v1.PromptXAIInvocationParameters,
            v1.PromptOllamaInvocationParameters,
            v1.PromptAwsInvocationParameters,
        ]
        if model_provider == "OPENAI":
            self._invocation_parameters = v1.PromptOpenAIInvocationParameters(
                type="openai",
                openai=v1.PromptOpenAIInvocationParametersContent(),
            )
        elif model_provider == "AZURE_OPENAI":
            self._invocation_parameters = v1.PromptAzureOpenAIInvocationParameters(
                type="azure_openai",
                azure_openai=v1.PromptAzureOpenAIInvocationParametersContent(),
            )
        elif model_provider == "ANTHROPIC":
            self._invocation_parameters = v1.PromptAnthropicInvocationParameters(
                type="anthropic",
                anthropic=v1.PromptAnthropicInvocationParametersContent(
                    max_tokens=1000,
                ),
            )
        elif model_provider == "GOOGLE":
            self._invocation_parameters = v1.PromptGoogleInvocationParameters(
                type="google",
                google=v1.PromptGoogleInvocationParametersContent(),
            )
        elif model_provider == "DEEPSEEK":
            self._invocation_parameters = v1.PromptDeepSeekInvocationParameters(
                type="deepseek",
                deepseek=v1.PromptDeepSeekInvocationParametersContent(),
            )
        elif model_provider == "XAI":
            self._invocation_parameters = v1.PromptXAIInvocationParameters(
                type="xai",
                xai=v1.PromptXAIInvocationParametersContent(),
            )
        elif model_provider == "OLLAMA":
            self._invocation_parameters = v1.PromptOllamaInvocationParameters(
                type="ollama",
                ollama=v1.PromptOllamaInvocationParametersContent(),
            )
        elif model_provider == "AWS":
            self._invocation_parameters = v1.PromptAwsInvocationParameters(
                type="aws",
                aws=v1.PromptAwsInvocationParametersContent(),
            )
        else:
            assert_never(model_provider)
        self._tools: Optional[v1.PromptTools] = None
        self._response_format: Optional[v1.PromptResponseFormatJSONSchema] = None
        self._id: Optional[str] = None

    def __dir__(self) -> list[str]:
        return [
            "id",
            "format",
            "from_openai",
            "from_anthropic",
        ]

    @property
    def id(self) -> Optional[str]:
        """
        Prompt Version ID if stored in the Phoenix backend
        """
        return self._id

    def format(
        self,
        *,
        variables: Mapping[str, str] = MappingProxyType({}),
        formatter: Optional[TemplateFormatter] = None,
        sdk: Optional[SDK] = None,
    ) -> _FormattedPrompt:
        """
        Formats the prompt for a specific SDK.

        Args:
            variables (Mapping[str, str]): A mapping of variable names to values to use in the
                prompt. Defaults to an empty mapping.
            formatter (Optional[TemplateFormatter]): A custom template formatter to use for the
                prompt. Defaults to None.
            sdk (Optional[SDK]): The SDK to format the prompt for. Defaults to None.

        Returns:
            The formatted prompt.
        """
        sdk = sdk or _to_sdk(self._model_provider)
        obj = self._dumps()
        if sdk == "openai":
            return OpenAIPrompt(
                *to_messages_openai(
                    obj,
                    variables=variables,
                    formatter=formatter,
                )
            )
        if sdk == "anthropic":
            return AnthropicPrompt(
                *to_messages_anthropic(
                    obj,
                    variables=variables,
                    formatter=formatter,
                )
            )
        if sdk == "google_generativeai":
            return GoogleGenerativeaiPrompt(
                *to_messages_google(
                    obj,
                    variables=variables,
                    formatter=formatter,
                )
            )
        if sdk == "boto3":
            raise NotImplementedError("Boto3 is not supported yet")
        assert_never(sdk)

    @classmethod
    def _loads(
        cls,
        obj: Union[v1.PromptVersionData, v1.PromptVersion],
    ) -> Self:
        assert obj["template"]["type"] == "chat"
        messages: Sequence[v1.PromptMessage] = obj["template"]["messages"]
        ans = cls(
            messages,
            model_name=obj["model_name"],
            description=obj.get("description"),
            model_provider=obj["model_provider"],
            template_format=obj["template_format"],
        )
        ans._invocation_parameters = obj["invocation_parameters"]
        if "tools" in obj:
            ans._tools = obj["tools"]
        if "response_format" in obj:
            ans._response_format = obj["response_format"]
        if "id" in obj:
            ans._id = obj["id"]  # type: ignore[typeddict-item]
        return ans

    def _dumps(
        self,
    ) -> v1.PromptVersionData:
        assert self._template["type"] == "chat"
        ans = v1.PromptVersionData(
            model_provider=self._model_provider,
            model_name=self._model_name,
            template=self._template,
            template_type=self._template_type,
            template_format=self._template_format,
            invocation_parameters=self._invocation_parameters,
        )
        if self._tools is not None:
            ans["tools"] = self._tools
        if self._response_format is not None:
            ans["response_format"] = self._response_format
        if self._description is not None:
            ans["description"] = self._description
        return ans

    @classmethod
    def from_openai(
        cls,
        obj: CompletionCreateParamsBase,
        /,
        *,
        template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = "MUSTACHE",
        description: Optional[str] = None,
        model_provider: Literal["OPENAI", "AZURE_OPENAI", "DEEPSEEK", "XAI", "OLLAMA"] = "OPENAI",
    ) -> Self:
        """
        Creates a prompt version from an OpenAI chat completion model.

        Args:
            obj (CompletionCreateParamsBase): The completion create parameters.
            template_format (Literal["F_STRING", "MUSTACHE", "NONE"]): The format of the template
                to use for the prompt. Defaults to "MUSTACHE".
            description (Optional[str]): A description of the prompt. Defaults to None.
            model_provider (Literal["OPENAI", "AZURE_OPENAI", "DEEPSEEK", "XAI", "OLLAMA"]):
                The provider of the model to use for the prompt. Defaults to "OPENAI".

        Returns:
            PromptVersion: The prompt version.
        """
        return cls._loads(
            create_prompt_version_from_openai(
                obj,
                description=description,
                template_format=template_format,
                model_provider=model_provider,
            )
        )

    @classmethod
    def from_aws(
        cls,
        obj: CompletionCreateParamsBase,
        /,
        *,
        template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = "MUSTACHE",
        description: Optional[str] = None,
        model_provider: Literal["AWS"] = "AWS",
    ) -> Self:
        raise NotImplementedError("AWS is not supported yet")

    @classmethod
    def from_anthropic(
        cls,
        obj: MessageCreateParamsBase,
        /,
        *,
        template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = "MUSTACHE",
        description: Optional[str] = None,
        model_provider: Literal["ANTHROPIC"] = "ANTHROPIC",
    ) -> Self:
        """
        Creates a prompt version from an Anthropic message model.

        Args:
            obj (MessageCreateParamsBase): The message create parameters.
            template_format (Literal["F_STRING", "MUSTACHE", "NONE"]): The format of the template
                to use for the prompt. Defaults to "MUSTACHE".
            description (Optional[str]): A description of the prompt. Defaults to None.
            model_provider (Literal["ANTHROPIC"]): The provider of the model to use for the prompt.
                Defaults to "ANTHROPIC".

        Returns:
            PromptVersion: The prompt version.
        """
        return cls._loads(
            create_prompt_version_from_anthropic(
                obj,
                description=description,
                template_format=template_format,
                model_provider=model_provider,
            )
        )

    @classmethod
    def from_google_generativeai(
        cls,
        obj: Any,
        /,
        *,
        template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = "MUSTACHE",
        description: Optional[str] = None,
        model_provider: Literal["GOOGLE"] = "GOOGLE",
    ) -> Self:
        return cls._loads(
            create_prompt_version_from_google(
                obj,
                description=description,
                template_format=template_format,
                model_provider=model_provider,
            )
        )


@dataclass(frozen=True)
class _FormattedPrompt(ABC, abc.Mapping[str, Any]):
    messages: Sequence[Any]
    kwargs: Mapping[str, Any]

    def __len__(self) -> int:
        return 1 + len(self.kwargs)

    def __iter__(self) -> abc.Iterator[str]:
        yield "messages"
        yield from self.kwargs

    def __getitem__(self, key: str) -> Any:
        if key == "messages":
            return self.messages
        return self.kwargs[key]


@dataclass(frozen=True)
class OpenAIPrompt(_FormattedPrompt):
    """
    Represents a formatted prompt for OpenAI chat completion models.

    Attributes:
        messages (Sequence[ChatCompletionMessageParam]): A sequence of chat completion message
            parameters.
        kwargs (OpenAIChatCompletionModelKwargs): Keyword arguments specific to OpenAI chat
            completion model invocation.
    """

    messages: Sequence[ChatCompletionMessageParam]
    kwargs: OpenAIChatCompletionModelKwargs


@dataclass(frozen=True)
class AnthropicPrompt(_FormattedPrompt):
    """
    Represents a formatted prompt for Anthropic message models.

    Attributes:
        messages (Sequence[MessageParam]): A sequence of message parameters.
        kwargs (AnthropicMessageModelKwargs): Keyword arguments specific to Anthropic message model
            invocation.
    """

    messages: Sequence[MessageParam]
    kwargs: AnthropicMessageModelKwargs


@dataclass(frozen=True)
class GoogleGenerativeaiPrompt(_FormattedPrompt):
    messages: Sequence[protos.Content]
    kwargs: GoogleModelKwargs


SDK: TypeAlias = Literal[
    "anthropic",  # https://pypi.org/project/anthropic/
    "google_generativeai",  # https://pypi.org/project/google-generativeai/
    "openai",  # https://pypi.org/project/openai/
    "boto3",  # https://boto3.amazonaws.com/v1/documentation/api/latest/index.html
]


def _to_sdk(
    model_provider: Literal[
        "OPENAI",
        "AZURE_OPENAI",
        "ANTHROPIC",
        "GOOGLE",
        "DEEPSEEK",
        "XAI",
        "OLLAMA",
        "AWS",
    ],
) -> SDK:
    if model_provider == "OPENAI":
        return "openai"
    if model_provider == "AZURE_OPENAI":
        return "openai"
    if model_provider == "ANTHROPIC":
        return "anthropic"
    if model_provider == "GOOGLE":
        return "google_generativeai"
    if model_provider == "DEEPSEEK":
        return "openai"
    if model_provider == "XAI":
        return "openai"
    if model_provider == "OLLAMA":
        return "openai"
    if model_provider == "AWS":
        return "boto3"
    assert_never(model_provider)



import gzip
from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import Any
from unittest.mock import Mock, mock_open, patch

import pandas as pd
import pytest

from phoenix.client.__generated__ import v1
from phoenix.client.resources.datasets import (
    Dataset,
    DatasetKeys,
    DatasetUploadError,
    _get_csv_column_headers,  # pyright: ignore[reportPrivateUsage]
    _infer_keys,  # pyright: ignore[reportPrivateUsage]
    _is_all_dict,  # pyright: ignore[reportPrivateUsage]
    _is_valid_dataset_example,  # pyright: ignore[reportPrivateUsage]
    _parse_datetime,  # pyright: ignore[reportPrivateUsage]
    _prepare_csv,  # pyright: ignore[reportPrivateUsage]
    _prepare_dataframe_as_csv,  # pyright: ignore[reportPrivateUsage]
)


class TestDataset:
    @pytest.fixture
    def dataset_info(self) -> v1.DatasetWithExampleCount:
        return v1.DatasetWithExampleCount(
            id="dataset123",
            name="Test Dataset",
            description="A test dataset",
            metadata={"key": "value"},
            created_at="2024-01-15T10:00:00",
            updated_at="2024-01-15T11:00:00",
            example_count=2,
        )

    @pytest.fixture
    def examples_data(self) -> v1.ListDatasetExamplesData:
        return v1.ListDatasetExamplesData(
            dataset_id="dataset123",
            version_id="version456",
            examples=[
                v1.DatasetExample(
                    id="ex1",
                    input={"text": "hello"},
                    output={"response": "hi"},
                    metadata={"source": "test"},
                    updated_at="2024-01-15T10:00:00",
                ),
                v1.DatasetExample(
                    id="ex2",
                    input={"text": "world"},
                    output={"response": "earth"},
                    metadata={"source": "test"},
                    updated_at="2024-01-15T10:00:00",
                ),
            ],
        )

    def test_properties(self, dataset_info: Any, examples_data: Any) -> None:
        dataset = Dataset(dataset_info, examples_data)

        assert dataset.id == "dataset123"
        assert dataset.name == "Test Dataset"
        assert dataset.description == "A test dataset"
        assert dataset.version_id == "version456"
        assert len(dataset.examples) == 2
        assert dataset.metadata == {"key": "value"}
        assert dataset.example_count == 2
        assert isinstance(dataset.created_at, datetime)
        assert isinstance(dataset.updated_at, datetime)

    def test_sequence_operations(self, dataset_info: Any, examples_data: Any) -> None:
        dataset = Dataset(dataset_info, examples_data)

        assert len(dataset) == 2
        assert dataset[0]["id"] == "ex1"
        assert dataset[1]["id"] == "ex2"

        examples = list(dataset)
        assert len(examples) == 2
        assert all(isinstance(ex, dict) for ex in examples)

    def test_to_dataframe(self, dataset_info: Any, examples_data: Any) -> None:
        dataset = Dataset(dataset_info, examples_data)
        df = dataset.to_dataframe()

        assert isinstance(df, pd.DataFrame)
        assert len(df) == 2
        assert list(df.columns) == ["input", "output", "metadata"]
        assert df.index.name == "example_id"  # pyright: ignore[reportUnknownMemberType]
        assert list(df.index) == ["ex1", "ex2"]  # pyright: ignore[reportUnknownMemberType,reportUnknownArgumentType]

    def test_to_dataframe_no_pandas(self, dataset_info: Any, examples_data: Any) -> None:
        dataset = Dataset(dataset_info, examples_data)

        with patch("builtins.__import__", side_effect=ImportError("No pandas")):
            with pytest.raises(ImportError, match="pandas is required"):
                dataset.to_dataframe()

    def test_repr(self, dataset_info: Any, examples_data: Any) -> None:
        dataset = Dataset(dataset_info, examples_data)
        repr_str = repr(dataset)
        assert "Dataset(" in repr_str
        assert "id='dataset123'" in repr_str
        assert "name='Test Dataset'" in repr_str


class TestDatasetKeys:
    def test_valid_keys(self) -> None:
        keys = DatasetKeys(
            input_keys=frozenset(["a", "b"]),
            output_keys=frozenset(["c"]),
            metadata_keys=frozenset(["d", "e"]),
        )
        assert keys.input == frozenset(["a", "b"])
        assert keys.output == frozenset(["c"])
        assert keys.metadata == frozenset(["d", "e"])
        assert set(keys) == {"a", "b", "c", "d", "e"}

    def test_overlapping_keys_validation(self) -> None:
        with pytest.raises(ValueError, match="Input and output keys overlap"):
            DatasetKeys(
                input_keys=frozenset(["a", "b"]),
                output_keys=frozenset(["b", "c"]),
                metadata_keys=frozenset(["d"]),
            )

    def test_check_differences(self) -> None:
        keys = DatasetKeys(
            input_keys=frozenset(["a"]),
            output_keys=frozenset(["b"]),
            metadata_keys=frozenset(["c"]),
        )

        keys.check_differences(frozenset(["a", "b", "c", "d"]))

        with pytest.raises(ValueError, match="Keys not found"):
            keys.check_differences(frozenset(["a", "b"]))


class TestHelperFunctions:
    def test_parse_datetime(self) -> None:
        dt = _parse_datetime("2024-01-15T10:30:00")
        assert dt.year == 2024
        assert dt.month == 1
        assert dt.day == 15
        assert dt.hour == 10
        assert dt.minute == 30

    def test_is_all_dict(self) -> None:
        assert _is_all_dict([{"a": 1}, {"b": 2}])
        assert not _is_all_dict([{"a": 1}, "not a dict"])
        assert _is_all_dict([])

    def test_is_valid_dataset_example(self) -> None:
        valid_example = v1.DatasetExample(
            id="ex1",
            input={"text": "hello"},
            output={"response": "hi"},
            metadata={"source": "test"},
            updated_at="2024-01-15T10:00:00",
        )
        assert _is_valid_dataset_example(valid_example)
        assert not _is_valid_dataset_example({"incomplete": "dict"})
        assert not _is_valid_dataset_example("not a dict")

    def test_get_csv_column_headers(self) -> None:
        csv_content = "col1,col2,col3\nval1,val2,val3\n"

        mock_path = Mock(spec=Path)
        mock_path.resolve.return_value = mock_path
        mock_path.is_file.return_value = True

        with patch("builtins.open", mock_open(read_data=csv_content)):
            headers = _get_csv_column_headers(mock_path)
            assert headers == ("col1", "col2", "col3")

    def test_get_csv_column_headers_errors(self) -> None:
        # Test empty file
        mock_path = Mock(spec=Path)
        mock_path.resolve.return_value = mock_path
        mock_path.is_file.return_value = True

        with patch("builtins.open", mock_open(read_data="")):
            with pytest.raises(ValueError, match="CSV file has no data rows"):
                _get_csv_column_headers(mock_path)

        # Test file not found
        mock_path.is_file.return_value = False
        with pytest.raises(FileNotFoundError):
            _get_csv_column_headers(mock_path)

    def test_infer_keys(self) -> None:
        # Test with DataFrame
        df = pd.DataFrame(
            {"input1": [1, 2], "input2": [3, 4], "response": [5, 6], "metadata1": [7, 8]}
        )

        input_keys, output_keys, metadata_keys = _infer_keys(df)

        assert input_keys == ("input1", "input2")
        assert output_keys == ("response",)
        assert metadata_keys == ("metadata1",)

        # Test with mocked CSV file by patching _get_csv_column_headers
        expected_headers = ("feature1", "feature2", "output", "extra")

        with patch(
            "phoenix.client.resources.datasets._get_csv_column_headers",
            return_value=expected_headers,
        ):
            input_keys, output_keys, metadata_keys = _infer_keys("/fake/path.csv")

            assert input_keys == ("feature1", "feature2")
            assert output_keys == ("output",)
            assert metadata_keys == ("extra",)

    def test_prepare_csv(self) -> None:
        csv_content = "input1,input2,output\nval1,val2,val3\n"
        mock_path = Mock(spec=Path)
        mock_path.resolve.return_value = mock_path
        mock_path.is_file.return_value = True
        mock_path.name = "test.csv"

        keys = DatasetKeys(
            input_keys=frozenset(["input1", "input2"]),
            output_keys=frozenset(["output"]),
            metadata_keys=frozenset(),
        )

        # Mock both text and binary file operations
        with patch("builtins.open", mock_open(read_data=csv_content)) as mock_file:
            # Configure mock to return bytes when opened in binary mode
            mock_file.return_value.read.return_value = csv_content.encode()

            name, file_obj, content_type, headers = _prepare_csv(mock_path, keys)

            assert name == "test.csv"
            assert content_type == "text/csv"
            assert headers == {"Content-Encoding": "gzip"}

            file_obj.seek(0)
            decompressed = gzip.decompress(file_obj.read()).decode()
            assert decompressed == csv_content

    def test_prepare_csv_validation(self) -> None:
        csv_content = "col1,col1,col2\nval1,val2,val3\n"
        mock_path = Mock(spec=Path)
        mock_path.resolve.return_value = mock_path
        mock_path.is_file.return_value = True

        keys = DatasetKeys(
            input_keys=frozenset(["col1"]),
            output_keys=frozenset(),
            metadata_keys=frozenset(),
        )

        with patch("builtins.open", mock_open(read_data=csv_content)):
            with pytest.raises(ValueError, match="Duplicate column headers"):
                _prepare_csv(mock_path, keys)

    def test_prepare_dataframe_as_json(self) -> None:
        df = pd.DataFrame({"input": ["a", "b"], "output": ["x", "y"], "metadata": ["m1", "m2"]})

        keys = DatasetKeys(
            input_keys=frozenset(["input"]),
            output_keys=frozenset(["output"]),
            metadata_keys=frozenset(["metadata"]),
        )

        name, file_obj, content_type, headers = _prepare_dataframe_as_csv(df, keys)

        assert name == "dataframe.csv"
        assert content_type == "text/csv"
        assert headers == {"Content-Encoding": "gzip"}

        file_obj.seek(0)
        decompressed = gzip.decompress(file_obj.read()).decode()
        assert "input,output,metadata" in decompressed  # CSV header
        assert "a,x,m1" in decompressed
        assert "b,y,m2" in decompressed

    def test_prepare_dataframe_validation(self) -> None:
        df = pd.DataFrame()
        keys = DatasetKeys(
            input_keys=frozenset(),
            output_keys=frozenset(),
            metadata_keys=frozenset(),
        )

        with pytest.raises(ValueError, match="DataFrame has no data"):
            _prepare_dataframe_as_csv(df, keys)


class TestDatasetUploadError:
    def test_error_message(self) -> None:
        error = DatasetUploadError("Upload failed")
        assert str(error) == "Upload failed"


class TestCSVProcessing:
    def test_csv_parsing_with_various_formats(self) -> None:
        # Test basic CSV
        basic_csv = "input,output,metadata\nhello,hi,test\nworld,earth,demo"
        df_basic = pd.read_csv(StringIO(basic_csv))  # pyright: ignore[reportUnknownMemberType]
        input_keys, output_keys, metadata_keys = _infer_keys(df_basic)
        assert input_keys == ("input",)
        assert output_keys == ("output",)
        assert metadata_keys == ("metadata",)

        # Test CSV with quoted fields containing commas
        quoted_csv = (
            'question,answer,notes\n"What is 2+2?","4","simple math"\n"What is the capital of '
            'France?","Paris","geography, Europe"'
        )
        df_quoted = pd.read_csv(StringIO(quoted_csv))  # pyright: ignore[reportUnknownMemberType]
        input_keys, output_keys, metadata_keys = _infer_keys(df_quoted)
        assert input_keys == ("question",)
        assert output_keys == ("answer",)
        assert metadata_keys == ("notes",)
        # Verify quoted field was parsed correctly
        assert df_quoted.iloc[1]["notes"] == "geography, Europe"  # pyright: ignore[reportGeneralTypeIssues]

        # Test CSV with response column detection
        response_csv = (
            "feature1,feature2,response,extra\nval1,val2,resp1,meta1\nval3,val4,resp2,meta2"
        )
        df_response = pd.read_csv(StringIO(response_csv))  # pyright: ignore[reportUnknownMemberType]
        input_keys, output_keys, metadata_keys = _infer_keys(df_response)
        assert input_keys == ("feature1", "feature2")
        assert output_keys == ("response",)
        assert metadata_keys == ("extra",)

    def test_csv_key_inference_patterns(self) -> None:
        """Test different patterns for output column detection."""
        test_cases = [
            ("input1,input2,output", ("input1", "input2"), ("output",), ()),
            ("feat1,feat2,answer", ("feat1", "feat2"), ("answer",), ()),
            ("x,y,response,meta", ("x", "y"), ("response",), ("meta",)),
            ("a,outputs,b", ("a",), ("outputs",), ("b",)),
            ("col1,col2", ("col1", "col2"), (), ()),  # No output column
        ]

        for csv_header, expected_input, expected_output, expected_metadata in test_cases:
            csv_content = (
                csv_header + "\nval1,val2" + (",val3" if csv_header.count(",") == 2 else "")
            )
            df = pd.read_csv(StringIO(csv_content))  # pyright: ignore[reportUnknownMemberType]
            input_keys, output_keys, metadata_keys = _infer_keys(df)
            assert input_keys == expected_input, f"Failed for {csv_header}"
            assert output_keys == expected_output, f"Failed for {csv_header}"
            assert metadata_keys == expected_metadata, f"Failed for {csv_header}"

    def test_dataframe_to_csv_preparation(self) -> None:
        # Create a DataFrame with complex data
        df = pd.DataFrame(
            {
                "input_text": ["What is AI?", "Define ML"],
                "input_context": ["technology", "computer science"],
                "output_answer": ["Artificial Intelligence", "Machine Learning"],
                "metadata_source": ["wiki", "textbook"],
            }
        )

        keys = DatasetKeys(
            input_keys=frozenset(["input_text", "input_context"]),
            output_keys=frozenset(["output_answer"]),
            metadata_keys=frozenset(["metadata_source"]),
        )

        name, file_obj, content_type, headers = _prepare_dataframe_as_csv(df, keys)

        assert name == "dataframe.csv"
        assert content_type == "text/csv"
        assert headers == {"Content-Encoding": "gzip"}

        file_obj.seek(0)
        decompressed = gzip.decompress(file_obj.read()).decode()

        generated_df = pd.read_csv(StringIO(decompressed))  # pyright: ignore[reportUnknownMemberType]

        # Verify structure - columns should be sorted alphabetically within each group
        # (input keys sorted, output keys sorted, metadata keys sorted)
        expected_columns = [
            "input_context",
            "input_text",
            "output_answer",
            "metadata_source",
        ]
        assert list(generated_df.columns) == expected_columns
        assert len(generated_df) == 2

        assert generated_df.iloc[0]["input_text"] == "What is AI?"  # pyright: ignore[reportGeneralTypeIssues]
        assert generated_df.iloc[1]["output_answer"] == "Machine Learning"  # pyright: ignore[reportGeneralTypeIssues]

    def test_csv_validation_edge_cases(self) -> None:
        """Test validation of problematic CSV scenarios."""
        df_with_duplicate_cols = pd.DataFrame([[1, 2, 3]], columns=["col1", "col1", "col2"])
        assert list(df_with_duplicate_cols.columns) == ["col1", "col1", "col2"]

        keys = DatasetKeys(
            input_keys=frozenset(["col1"]),
            output_keys=frozenset(["col2"]),
            metadata_keys=frozenset(),
        )

        with pytest.raises(ValueError, match="Duplicate column names in DataFrame"):
            _prepare_dataframe_as_csv(df_with_duplicate_cols, keys)

        df = pd.DataFrame({"col1": [1], "col2": [2]})
        keys_with_missing = DatasetKeys(
            input_keys=frozenset(["col1", "missing_col"]),
            output_keys=frozenset(["col2"]),
            metadata_keys=frozenset(),
        )

        with pytest.raises(ValueError, match="Keys not found"):
            keys_with_missing.check_differences(frozenset(df.columns))



"""Do not edit"""

from __future__ import annotations

from typing import Any, Literal, Mapping, Optional, Sequence, TypedDict, Union

from typing_extensions import NotRequired


class CategoricalAnnotationValue(TypedDict):
    label: str
    score: NotRequired[float]


class CreateExperimentRequestBody(TypedDict):
    name: NotRequired[str]
    description: NotRequired[str]
    metadata: NotRequired[Mapping[str, Any]]
    version_id: NotRequired[str]
    repetitions: NotRequired[int]


class CreateExperimentRunRequestBody(TypedDict):
    dataset_example_id: str
    output: Any
    repetition_number: int
    start_time: str
    end_time: str
    trace_id: NotRequired[str]
    error: NotRequired[str]


class CreateExperimentRunResponseBodyData(TypedDict):
    id: str


class CreateProjectRequestBody(TypedDict):
    name: str
    description: NotRequired[str]


class CreateSpansResponseBody(TypedDict):
    total_received: int
    total_queued: int


class Dataset(TypedDict):
    id: str
    name: str
    description: Optional[str]
    metadata: Mapping[str, Any]
    created_at: str
    updated_at: str


class DatasetExample(TypedDict):
    id: str
    input: Mapping[str, Any]
    output: Mapping[str, Any]
    metadata: Mapping[str, Any]
    updated_at: str


class DatasetVersion(TypedDict):
    version_id: str
    description: Optional[str]
    metadata: Mapping[str, Any]
    created_at: str


class DatasetWithExampleCount(TypedDict):
    id: str
    name: str
    description: Optional[str]
    metadata: Mapping[str, Any]
    created_at: str
    updated_at: str
    example_count: int


class Experiment(TypedDict):
    id: str
    dataset_id: str
    dataset_version_id: str
    repetitions: int
    metadata: Mapping[str, Any]
    project_name: Optional[str]
    created_at: str
    updated_at: str


class ExperimentEvaluationResult(TypedDict):
    label: NotRequired[str]
    score: NotRequired[float]
    explanation: NotRequired[str]


class ExperimentRunResponse(TypedDict):
    dataset_example_id: str
    output: Any
    repetition_number: int
    start_time: str
    end_time: str
    id: str
    experiment_id: str
    trace_id: NotRequired[str]
    error: NotRequired[str]


class FreeformAnnotationConfig(TypedDict):
    type: Literal["FREEFORM"]
    name: str
    id: str
    description: NotRequired[str]


class FreeformAnnotationConfigData(TypedDict):
    type: Literal["FREEFORM"]
    name: str
    description: NotRequired[str]


class GetDatasetResponseBody(TypedDict):
    data: DatasetWithExampleCount


class GetExperimentResponseBody(TypedDict):
    data: Experiment


class InsertedSpanAnnotation(TypedDict):
    id: str


class ListDatasetExamplesData(TypedDict):
    dataset_id: str
    version_id: str
    examples: Sequence[DatasetExample]


class ListDatasetExamplesResponseBody(TypedDict):
    data: ListDatasetExamplesData


class ListDatasetVersionsResponseBody(TypedDict):
    data: Sequence[DatasetVersion]
    next_cursor: Optional[str]


class ListDatasetsResponseBody(TypedDict):
    data: Sequence[Dataset]
    next_cursor: Optional[str]


class ListExperimentRunsResponseBody(TypedDict):
    data: Sequence[ExperimentRunResponse]


class ListExperimentsResponseBody(TypedDict):
    data: Sequence[Experiment]


class LocalUserData(TypedDict):
    email: str
    username: str
    role: Literal["SYSTEM", "ADMIN", "MEMBER"]
    auth_method: Literal["LOCAL"]
    password: NotRequired[str]


class LocalUser(LocalUserData):
    id: str
    created_at: str
    updated_at: str
    password_needs_reset: bool


class OAuth2UserData(TypedDict):
    email: str
    username: str
    role: Literal["SYSTEM", "ADMIN", "MEMBER"]
    auth_method: Literal["OAUTH2"]
    oauth2_client_id: NotRequired[str]
    oauth2_user_id: NotRequired[str]


class OAuth2User(OAuth2UserData):
    id: str
    created_at: str
    updated_at: str
    profile_picture_url: NotRequired[str]


class OtlpStatus(TypedDict):
    code: NotRequired[int]
    message: NotRequired[str]


class Project(TypedDict):
    name: str
    id: str
    description: NotRequired[str]


class PromptData(TypedDict):
    name: str
    description: NotRequired[str]
    source_prompt_id: NotRequired[str]


class Prompt(PromptData):
    id: str


class PromptAnthropicThinkingConfigDisabled(TypedDict):
    type: Literal["disabled"]


class PromptAnthropicThinkingConfigEnabled(TypedDict):
    type: Literal["enabled"]
    budget_tokens: int


class PromptAwsInvocationParametersContent(TypedDict):
    max_tokens: NotRequired[int]
    temperature: NotRequired[float]
    top_p: NotRequired[float]


class PromptAzureOpenAIInvocationParametersContent(TypedDict):
    temperature: NotRequired[float]
    max_tokens: NotRequired[int]
    max_completion_tokens: NotRequired[int]
    frequency_penalty: NotRequired[float]
    presence_penalty: NotRequired[float]
    top_p: NotRequired[float]
    seed: NotRequired[int]
    reasoning_effort: NotRequired[Literal["low", "medium", "high"]]


class PromptDeepSeekInvocationParametersContent(TypedDict):
    temperature: NotRequired[float]
    max_tokens: NotRequired[int]
    max_completion_tokens: NotRequired[int]
    frequency_penalty: NotRequired[float]
    presence_penalty: NotRequired[float]
    top_p: NotRequired[float]
    seed: NotRequired[int]
    reasoning_effort: NotRequired[Literal["low", "medium", "high"]]


class PromptGoogleInvocationParametersContent(TypedDict):
    temperature: NotRequired[float]
    max_output_tokens: NotRequired[int]
    stop_sequences: NotRequired[Sequence[str]]
    presence_penalty: NotRequired[float]
    frequency_penalty: NotRequired[float]
    top_p: NotRequired[float]
    top_k: NotRequired[int]


class PromptOllamaInvocationParametersContent(TypedDict):
    temperature: NotRequired[float]
    max_tokens: NotRequired[int]
    max_completion_tokens: NotRequired[int]
    frequency_penalty: NotRequired[float]
    presence_penalty: NotRequired[float]
    top_p: NotRequired[float]
    seed: NotRequired[int]
    reasoning_effort: NotRequired[Literal["low", "medium", "high"]]


class PromptOpenAIInvocationParametersContent(TypedDict):
    temperature: NotRequired[float]
    max_tokens: NotRequired[int]
    max_completion_tokens: NotRequired[int]
    frequency_penalty: NotRequired[float]
    presence_penalty: NotRequired[float]
    top_p: NotRequired[float]
    seed: NotRequired[int]
    reasoning_effort: NotRequired[Literal["low", "medium", "high"]]


class PromptResponseFormatJSONSchemaDefinition(TypedDict):
    name: str
    description: NotRequired[str]
    schema: NotRequired[Mapping[str, Any]]
    strict: NotRequired[bool]


class PromptStringTemplate(TypedDict):
    type: Literal["string"]
    template: str


class PromptToolChoiceNone(TypedDict):
    type: Literal["none"]


class PromptToolChoiceOneOrMore(TypedDict):
    type: Literal["one_or_more"]


class PromptToolChoiceSpecificFunctionTool(TypedDict):
    type: Literal["specific_function"]
    function_name: str


class PromptToolChoiceZeroOrMore(TypedDict):
    type: Literal["zero_or_more"]


class PromptToolFunctionDefinition(TypedDict):
    name: str
    description: NotRequired[str]
    parameters: NotRequired[Mapping[str, Any]]
    strict: NotRequired[bool]


class PromptVersionTag(TypedDict):
    name: str
    id: str
    description: NotRequired[str]


class PromptVersionTagData(TypedDict):
    name: str
    description: NotRequired[str]


class PromptXAIInvocationParametersContent(TypedDict):
    temperature: NotRequired[float]
    max_tokens: NotRequired[int]
    max_completion_tokens: NotRequired[int]
    frequency_penalty: NotRequired[float]
    presence_penalty: NotRequired[float]
    top_p: NotRequired[float]
    seed: NotRequired[int]
    reasoning_effort: NotRequired[Literal["low", "medium", "high"]]


class SpanAnnotationResult(TypedDict):
    label: NotRequired[str]
    score: NotRequired[float]
    explanation: NotRequired[str]


class SpanContext(TypedDict):
    trace_id: str
    span_id: str


class SpanEvent(TypedDict):
    name: str
    timestamp: str
    attributes: NotRequired[Mapping[str, Any]]


class TextContentPart(TypedDict):
    type: Literal["text"]
    text: str


class ToolCallFunction(TypedDict):
    type: Literal["function"]
    name: str
    arguments: str


class ToolResultContentPart(TypedDict):
    type: Literal["tool_result"]
    tool_call_id: str
    tool_result: Optional[Union[bool, int, float, str, Mapping[str, Any], Sequence[Any]]]


class UpdateProjectRequestBody(TypedDict):
    description: NotRequired[str]


class UpdateProjectResponseBody(TypedDict):
    data: Project


class UploadDatasetData(TypedDict):
    dataset_id: str
    version_id: str


class UploadDatasetResponseBody(TypedDict):
    data: UploadDatasetData


class UpsertExperimentEvaluationRequestBody(TypedDict):
    experiment_run_id: str
    name: str
    annotator_kind: Literal["LLM", "CODE", "HUMAN"]
    start_time: str
    end_time: str
    result: ExperimentEvaluationResult
    error: NotRequired[str]
    metadata: NotRequired[Mapping[str, Any]]
    trace_id: NotRequired[str]


class UpsertExperimentEvaluationResponseBodyData(TypedDict):
    id: str


class ValidationError(TypedDict):
    loc: Sequence[Union[str, int]]
    msg: str
    type: str


class AnnotateSpansResponseBody(TypedDict):
    data: Sequence[InsertedSpanAnnotation]


class CategoricalAnnotationConfig(TypedDict):
    type: Literal["CATEGORICAL"]
    name: str
    optimization_direction: Literal["MINIMIZE", "MAXIMIZE", "NONE"]
    values: Sequence[CategoricalAnnotationValue]
    id: str
    description: NotRequired[str]


class CategoricalAnnotationConfigData(TypedDict):
    type: Literal["CATEGORICAL"]
    name: str
    optimization_direction: Literal["MINIMIZE", "MAXIMIZE", "NONE"]
    values: Sequence[CategoricalAnnotationValue]
    description: NotRequired[str]


class ContinuousAnnotationConfig(TypedDict):
    type: Literal["CONTINUOUS"]
    name: str
    optimization_direction: Literal["MINIMIZE", "MAXIMIZE", "NONE"]
    id: str
    description: NotRequired[str]
    lower_bound: NotRequired[float]
    upper_bound: NotRequired[float]


class ContinuousAnnotationConfigData(TypedDict):
    type: Literal["CONTINUOUS"]
    name: str
    optimization_direction: Literal["MINIMIZE", "MAXIMIZE", "NONE"]
    description: NotRequired[str]
    lower_bound: NotRequired[float]
    upper_bound: NotRequired[float]


class CreateAnnotationConfigResponseBody(TypedDict):
    data: Union[CategoricalAnnotationConfig, ContinuousAnnotationConfig, FreeformAnnotationConfig]


class CreateExperimentResponseBody(TypedDict):
    data: Experiment


class CreateExperimentRunResponseBody(TypedDict):
    data: CreateExperimentRunResponseBodyData


class CreateProjectResponseBody(TypedDict):
    data: Project


class CreateUserRequestBody(TypedDict):
    user: Union[LocalUserData, OAuth2UserData]
    send_welcome_email: NotRequired[bool]


class CreateUserResponseBody(TypedDict):
    data: Union[LocalUser, OAuth2User]


class DeleteAnnotationConfigResponseBody(TypedDict):
    data: Union[CategoricalAnnotationConfig, ContinuousAnnotationConfig, FreeformAnnotationConfig]


class GetAnnotationConfigResponseBody(TypedDict):
    data: Union[CategoricalAnnotationConfig, ContinuousAnnotationConfig, FreeformAnnotationConfig]


class GetAnnotationConfigsResponseBody(TypedDict):
    data: Sequence[
        Union[CategoricalAnnotationConfig, ContinuousAnnotationConfig, FreeformAnnotationConfig]
    ]
    next_cursor: Optional[str]


class GetProjectResponseBody(TypedDict):
    data: Project


class GetProjectsResponseBody(TypedDict):
    data: Sequence[Project]
    next_cursor: Optional[str]


class GetPromptVersionTagsResponseBody(TypedDict):
    data: Sequence[PromptVersionTag]
    next_cursor: Optional[str]


class GetPromptsResponseBody(TypedDict):
    data: Sequence[Prompt]
    next_cursor: Optional[str]


class GetUsersResponseBody(TypedDict):
    data: Sequence[Union[LocalUser, OAuth2User]]
    next_cursor: Optional[str]


class HTTPValidationError(TypedDict):
    detail: NotRequired[Sequence[ValidationError]]


class PromptAnthropicInvocationParametersContent(TypedDict):
    max_tokens: int
    temperature: NotRequired[float]
    top_p: NotRequired[float]
    stop_sequences: NotRequired[Sequence[str]]
    thinking: NotRequired[
        Union[PromptAnthropicThinkingConfigDisabled, PromptAnthropicThinkingConfigEnabled]
    ]


class PromptAwsInvocationParameters(TypedDict):
    type: Literal["aws"]
    aws: PromptAwsInvocationParametersContent


class PromptAzureOpenAIInvocationParameters(TypedDict):
    type: Literal["azure_openai"]
    azure_openai: PromptAzureOpenAIInvocationParametersContent


class PromptDeepSeekInvocationParameters(TypedDict):
    type: Literal["deepseek"]
    deepseek: PromptDeepSeekInvocationParametersContent


class PromptGoogleInvocationParameters(TypedDict):
    type: Literal["google"]
    google: PromptGoogleInvocationParametersContent


class PromptOllamaInvocationParameters(TypedDict):
    type: Literal["ollama"]
    ollama: PromptOllamaInvocationParametersContent


class PromptOpenAIInvocationParameters(TypedDict):
    type: Literal["openai"]
    openai: PromptOpenAIInvocationParametersContent


class PromptResponseFormatJSONSchema(TypedDict):
    type: Literal["json_schema"]
    json_schema: PromptResponseFormatJSONSchemaDefinition


class PromptToolFunction(TypedDict):
    type: Literal["function"]
    function: PromptToolFunctionDefinition


class PromptTools(TypedDict):
    type: Literal["tools"]
    tools: Sequence[PromptToolFunction]
    tool_choice: NotRequired[
        Union[
            PromptToolChoiceNone,
            PromptToolChoiceZeroOrMore,
            PromptToolChoiceOneOrMore,
            PromptToolChoiceSpecificFunctionTool,
        ]
    ]
    disable_parallel_tool_calls: NotRequired[bool]


class PromptXAIInvocationParameters(TypedDict):
    type: Literal["xai"]
    xai: PromptXAIInvocationParametersContent


class Span(TypedDict):
    name: str
    context: SpanContext
    span_kind: str
    start_time: str
    end_time: str
    status_code: str
    id: NotRequired[str]
    parent_id: NotRequired[str]
    status_message: NotRequired[str]
    attributes: NotRequired[Mapping[str, Any]]
    events: NotRequired[Sequence[SpanEvent]]


class SpanAnnotationData(TypedDict):
    span_id: str
    name: str
    annotator_kind: Literal["LLM", "CODE", "HUMAN"]
    result: NotRequired[SpanAnnotationResult]
    metadata: NotRequired[Mapping[str, Any]]
    identifier: NotRequired[str]


class SpanAnnotation(SpanAnnotationData):
    id: str
    created_at: str
    updated_at: str
    source: Literal["API", "APP"]
    user_id: Optional[str]


class SpanAnnotationsResponseBody(TypedDict):
    data: Sequence[SpanAnnotation]
    next_cursor: Optional[str]


class SpansResponseBody(TypedDict):
    data: Sequence[Span]
    next_cursor: Optional[str]


class ToolCallContentPart(TypedDict):
    type: Literal["tool_call"]
    tool_call_id: str
    tool_call: ToolCallFunction


class UpdateAnnotationConfigResponseBody(TypedDict):
    data: Union[CategoricalAnnotationConfig, ContinuousAnnotationConfig, FreeformAnnotationConfig]


class UpsertExperimentEvaluationResponseBody(TypedDict):
    data: UpsertExperimentEvaluationResponseBodyData


class AnnotateSpansRequestBody(TypedDict):
    data: Sequence[SpanAnnotationData]


class CreateSpansRequestBody(TypedDict):
    data: Sequence[Span]


class PromptAnthropicInvocationParameters(TypedDict):
    type: Literal["anthropic"]
    anthropic: PromptAnthropicInvocationParametersContent


class PromptMessage(TypedDict):
    role: Literal["user", "assistant", "model", "ai", "tool", "system", "developer"]
    content: Union[
        str, Sequence[Union[TextContentPart, ToolCallContentPart, ToolResultContentPart]]
    ]


class PromptChatTemplate(TypedDict):
    type: Literal["chat"]
    messages: Sequence[PromptMessage]


class PromptVersionData(TypedDict):
    model_provider: Literal[
        "OPENAI", "AZURE_OPENAI", "ANTHROPIC", "GOOGLE", "DEEPSEEK", "XAI", "OLLAMA", "AWS"
    ]
    model_name: str
    template: Union[PromptChatTemplate, PromptStringTemplate]
    template_type: Literal["STR", "CHAT"]
    template_format: Literal["MUSTACHE", "F_STRING", "NONE"]
    invocation_parameters: Union[
        PromptOpenAIInvocationParameters,
        PromptAzureOpenAIInvocationParameters,
        PromptAnthropicInvocationParameters,
        PromptGoogleInvocationParameters,
        PromptDeepSeekInvocationParameters,
        PromptXAIInvocationParameters,
        PromptOllamaInvocationParameters,
        PromptAwsInvocationParameters,
    ]
    description: NotRequired[str]
    tools: NotRequired[PromptTools]
    response_format: NotRequired[PromptResponseFormatJSONSchema]


class PromptVersion(PromptVersionData):
    id: str


class CreatePromptRequestBody(TypedDict):
    prompt: PromptData
    version: PromptVersionData


class CreatePromptResponseBody(TypedDict):
    data: PromptVersion


class GetPromptResponseBody(TypedDict):
    data: PromptVersion


class GetPromptVersionsResponseBody(TypedDict):
    data: Sequence[PromptVersion]
    next_cursor: Optional[str]


class OtlpAnyValue(TypedDict):
    array_value: NotRequired[OtlpArrayValue]
    bool_value: NotRequired[bool]
    bytes_value: NotRequired[str]
    double_value: NotRequired[Union[float, str, Literal["Infinity", "-Infinity", "NaN"]]]
    int_value: NotRequired[Union[int, str]]
    kvlist_value: NotRequired[None]
    string_value: NotRequired[str]


class OtlpArrayValue(TypedDict):
    values: NotRequired[Sequence[OtlpAnyValue]]


class OtlpEvent(TypedDict):
    attributes: NotRequired[Sequence[OtlpKeyValue]]
    dropped_attributes_count: NotRequired[int]
    name: NotRequired[str]
    time_unix_nano: NotRequired[Union[int, str]]


class OtlpKeyValue(TypedDict):
    key: NotRequired[str]
    value: NotRequired[OtlpAnyValue]


class OtlpSpan(TypedDict):
    attributes: NotRequired[Sequence[OtlpKeyValue]]
    dropped_attributes_count: NotRequired[int]
    dropped_events_count: NotRequired[int]
    dropped_links_count: NotRequired[int]
    end_time_unix_nano: NotRequired[Union[int, str]]
    events: NotRequired[Sequence[OtlpEvent]]
    flags: NotRequired[int]
    kind: NotRequired[
        Union[
            int,
            Literal[
                "SPAN_KIND_UNSPECIFIED",
                "SPAN_KIND_INTERNAL",
                "SPAN_KIND_SERVER",
                "SPAN_KIND_CLIENT",
                "SPAN_KIND_PRODUCER",
                "SPAN_KIND_CONSUMER",
            ],
        ]
    ]
    links: NotRequired[None]
    name: NotRequired[str]
    parent_span_id: NotRequired[str]
    span_id: NotRequired[str]
    start_time_unix_nano: NotRequired[Union[int, str]]
    status: NotRequired[OtlpStatus]
    trace_id: NotRequired[str]
    trace_state: NotRequired[str]


class OtlpSpansResponseBody(TypedDict):
    data: Sequence[OtlpSpan]
    next_cursor: Optional[str]



from __future__ import annotations

import json
from enum import Enum
from random import randint, random
from secrets import token_hex
from typing import TYPE_CHECKING, Any, Iterable, Mapping, Optional, Union, cast

import pytest
from deepdiff.diff import DeepDiff
from faker import Faker
from openai.lib._parsing import type_to_response_format_param
from openai.lib._tools import pydantic_function_tool
from pydantic import BaseModel, create_model

from phoenix.client.__generated__ import v1
from phoenix.client.helpers.sdk.openai.chat import (
    _FunctionToolConversion,
    _MessageConversion,
    _ResponseFormatConversion,
    _TextContentPartConversion,
    _ToolCallContentPartConversion,
    _ToolKwargs,
    _ToolKwargsConversion,
    create_prompt_version_from_openai,
    to_chat_messages_and_kwargs,
)
from phoenix.client.utils.template_formatters import NO_OP_FORMATTER

if TYPE_CHECKING:
    from openai.types.chat import (
        ChatCompletionAssistantMessageParam,
        ChatCompletionContentPartParam,
        ChatCompletionContentPartTextParam,
        ChatCompletionDeveloperMessageParam,
        ChatCompletionMessageToolCallParam,
        ChatCompletionSystemMessageParam,
        ChatCompletionToolMessageParam,
        ChatCompletionToolParam,
        ChatCompletionUserMessageParam,
    )
    from openai.types.chat.chat_completion_assistant_message_param import ContentArrayOfContentPart
    from openai.types.chat.completion_create_params import (
        CompletionCreateParamsBase,
        ResponseFormat,
    )

fake = Faker()


def _dict() -> dict[str, Any]:
    return fake.pydict(3, value_types=(int, float, bool, str))  # pyright: ignore[reportUnknownMemberType]


def _str() -> str:
    return fake.pystr(8, 8)


def _user_msg(
    content: Union[str, Iterable[ChatCompletionContentPartParam]],
) -> ChatCompletionUserMessageParam:
    return {
        "role": "user",
        "content": content,
    }


def _assistant_msg(
    content: Optional[Union[str, Iterable[ContentArrayOfContentPart]]] = None,
    tool_calls: Iterable[ChatCompletionMessageToolCallParam] = (),
) -> ChatCompletionAssistantMessageParam:
    if not tool_calls:
        return {
            "role": "assistant",
            "content": content,
        }
    if content is None:
        return {
            "role": "assistant",
            "tool_calls": tool_calls,
        }
    return {
        "role": "assistant",
        "content": content,
        "tool_calls": tool_calls,
    }


def _tool_msg(
    content: Union[str, Iterable[ChatCompletionContentPartTextParam]],
) -> ChatCompletionToolMessageParam:
    return {
        "role": "tool",
        "content": content,
        "tool_call_id": _str(),
    }


def _system_msg(
    content: Union[str, Iterable[ChatCompletionContentPartTextParam]],
) -> ChatCompletionSystemMessageParam:
    return {
        "role": "system",
        "content": content,
    }


def _developer_msg(
    content: Union[str, Iterable[ChatCompletionContentPartTextParam]],
) -> ChatCompletionDeveloperMessageParam:
    return {
        "role": "developer",
        "content": content,
    }


def _text() -> ChatCompletionContentPartTextParam:
    return {
        "type": "text",
        "text": _str(),
    }


def _tool_call() -> ChatCompletionMessageToolCallParam:
    return {
        "id": _str(),
        "type": "function",
        "function": {"name": _str(), "arguments": json.dumps(_dict())},
    }


def _tool(name: Optional[str] = None) -> ChatCompletionToolParam:
    strict = fake.pybool()
    return {
        "type": "function",
        "function": {
            "name": name or _str(),
            "description": _str(),
            "parameters": {
                "type": "object",
                "properties": {
                    "x": {"type": "int", "description": _str()},
                    "y": {"type": "string", "description": _str()},
                },
                "required": ["x", "y"],
                "additionalProperties": strict or fake.pybool(),
            },
            "strict": strict,
        },
    }


class TestChatCompletionUserMessageParam:
    @pytest.mark.parametrize(
        "obj",
        [
            _user_msg(_str()),
            _user_msg([_text(), _text()]),
        ],
    )
    def test_round_trip(self, obj: ChatCompletionUserMessageParam) -> None:
        x: v1.PromptMessage = _MessageConversion.from_openai(obj)
        assert not DeepDiff(obj, next(_MessageConversion.to_openai(x, {}, NO_OP_FORMATTER)))


class TestChatCompletionSystemMessageParam:
    @pytest.mark.parametrize(
        "obj",
        [
            _system_msg(_str()),
            _system_msg([_text(), _text()]),
        ],
    )
    def test_round_trip(self, obj: ChatCompletionSystemMessageParam) -> None:
        x: v1.PromptMessage = _MessageConversion.from_openai(obj)
        assert not DeepDiff([obj], list(_MessageConversion.to_openai(x, {}, NO_OP_FORMATTER)))


class TestChatCompletionDeveloperMessageParam:
    @pytest.mark.parametrize(
        "obj",
        [
            _developer_msg(_str()),
            _developer_msg([_text(), _text()]),
        ],
    )
    def test_round_trip(self, obj: ChatCompletionSystemMessageParam) -> None:
        x: v1.PromptMessage = _MessageConversion.from_openai(obj)
        assert not DeepDiff([obj], list(_MessageConversion.to_openai(x, {}, NO_OP_FORMATTER)))


class TestChatCompletionAssistantMessageParam:
    @pytest.mark.parametrize(
        "obj",
        [
            _assistant_msg(_str()),
            _assistant_msg([_text(), _text()]),
            _assistant_msg(None, [_tool_call(), _tool_call()]),
            _assistant_msg(_str(), [_tool_call(), _tool_call()]),
            _assistant_msg([_text(), _text()], [_tool_call(), _tool_call()]),
        ],
    )
    def test_round_trip(self, obj: ChatCompletionAssistantMessageParam) -> None:
        x: v1.PromptMessage = _MessageConversion.from_openai(obj)
        assert not DeepDiff([obj], list(_MessageConversion.to_openai(x, {}, NO_OP_FORMATTER)))


class TestChatCompletionToolMessageParam:
    @pytest.mark.parametrize(
        "obj",
        [
            _tool_msg(_str()),
            _tool_msg([_text(), _text()]),
        ],
    )
    def test_round_trip(self, obj: ChatCompletionToolMessageParam) -> None:
        x: v1.PromptMessage = _MessageConversion.from_openai(obj)
        assert not DeepDiff([obj], list(_MessageConversion.to_openai(x, {}, NO_OP_FORMATTER)))


class TestFunctionToolConversion:
    @pytest.mark.parametrize(
        "obj",
        [_tool()],
    )
    def test_round_trip(self, obj: ChatCompletionToolParam) -> None:
        new_obj = _FunctionToolConversion.to_openai(_FunctionToolConversion.from_openai(obj))
        assert not DeepDiff(obj, new_obj)


class TestTextContentPartConversion:
    def test_round_trip(self) -> None:
        obj: ChatCompletionContentPartTextParam = _text()
        x: v1.TextContentPart = _TextContentPartConversion.from_openai(obj)
        new_obj: ChatCompletionContentPartTextParam = _TextContentPartConversion.to_openai(
            x, {}, NO_OP_FORMATTER
        )
        assert not DeepDiff(obj, new_obj)

    def test_formatter(self) -> None:
        x = v1.TextContentPart(type="text", text=_str())
        formatter, variables = _MockFormatter(), _dict()
        ans: ChatCompletionContentPartTextParam = _TextContentPartConversion.to_openai(
            x, variables, formatter
        )
        assert ans["text"] == formatter.format(x["text"], variables=variables)


class _GetWeather(BaseModel):
    city: str


class _GetPopulation(BaseModel):
    country: str
    year: Optional[int] = None


_TOOLS: list[ChatCompletionToolParam] = [
    cast(
        "ChatCompletionToolParam",
        json.loads(json.dumps(pydantic_function_tool(t))),
    )
    for t in cast(Iterable[type[BaseModel]], [_GetWeather, _GetPopulation])
]


class TestToolCallContentPartConversion:
    def test_round_trip(self) -> None:
        obj: ChatCompletionMessageToolCallParam = _tool_call()
        x: v1.ToolCallContentPart = _ToolCallContentPartConversion.from_openai(obj)
        new_obj: ChatCompletionMessageToolCallParam = _ToolCallContentPartConversion.to_openai(
            x, {}, NO_OP_FORMATTER
        )
        assert not DeepDiff(obj, new_obj)


class _UIType(str, Enum):
    div = "div"
    button = "button"
    header = "header"
    section = "section"
    field = "field"
    form = "form"


class _Attribute(BaseModel):
    name: str
    value: str


class _UI(BaseModel):
    type: _UIType
    label: str
    children: list[_UI]
    attributes: list[_Attribute]


_UI.model_rebuild()


class TestResponseFormatJSONSchemaConversion:
    @pytest.mark.parametrize(
        "type_",
        [
            create_model("Response", ui=(_UI, ...)),
        ],
    )
    def test_round_trip(self, type_: type[BaseModel]) -> None:
        obj = cast("ResponseFormat", type_to_response_format_param(type_))
        x: v1.PromptResponseFormatJSONSchema = _ResponseFormatConversion.from_openai(obj)
        new_obj = _ResponseFormatConversion.to_openai(x)
        assert not DeepDiff(obj, new_obj)


class TestToolKwargsConversion:
    @pytest.mark.parametrize(
        "obj",
        [
            {},
            {
                "tools": [_tool(), _tool()],
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": "none",
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": "none",
                "parallel_tool_calls": False,
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": "auto",
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": "auto",
                "parallel_tool_calls": False,
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": "required",
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": "required",
                "parallel_tool_calls": False,
            },
            {
                "tools": [_tool(), _tool("xyz")],
                "tool_choice": {
                    "type": "function",
                    "function": {"name": "xyz"},
                },
            },
            {
                "tools": [_tool(), _tool("xyz")],
                "tool_choice": {
                    "type": "function",
                    "function": {"name": "xyz"},
                },
                "parallel_tool_calls": False,
            },
        ],
    )
    def test_round_trip(self, obj: _ToolKwargs) -> None:
        x: Optional[v1.PromptTools] = _ToolKwargsConversion.from_openai(obj)
        new_obj: _ToolKwargs = _ToolKwargsConversion.to_openai(x)
        assert not DeepDiff(obj, new_obj)


class TestCompletionCreateParamsBase:
    @pytest.mark.parametrize(
        "obj",
        [
            {
                "model": token_hex(8),
                "messages": [
                    {
                        "role": "system",
                        "content": "You will be provided with statements, and your task is"
                        "to convert them to standard English.",
                    },
                    {"role": "user", "content": "{{ statement }}"},
                ],
                "temperature": random(),
                "max_completion_tokens": randint(1, 256),
                "top_p": random(),
            },
            {
                "model": token_hex(8),
                "messages": [
                    {
                        "role": "system",
                        "content": "You are a UI generator. Convert the user input into a UI.",
                    },
                    {
                        "role": "user",
                        "content": "Make a form for {{ feature }}.",
                    },
                ],
                "response_format": cast(
                    "ResponseFormat",
                    type_to_response_format_param(
                        create_model("Response", ui=(_UI, ...)),
                    ),
                ),
                "temperature": random(),
                "max_completion_tokens": randint(1, 256),
                "top_p": random(),
            },
            {
                "model": token_hex(8),
                "messages": [
                    {
                        "role": "user",
                        "content": "What is the latest population estimate for {{ location }}?",
                    }
                ],
                "tools": _TOOLS,
                "tool_choice": "required",
                "temperature": random(),
                "max_completion_tokens": randint(1, 256),
                "top_p": random(),
            },
            {
                "model": token_hex(8),
                "messages": [
                    {
                        "role": "system",
                        "content": "You are a UI generator. Convert the user input into a UI.",
                    },
                    {
                        "role": "user",
                        "content": "Make a form for {{ feature }}.",
                    },
                ],
                "response_format": cast(
                    "ResponseFormat",
                    type_to_response_format_param(
                        create_model("Response", ui=(_UI, ...)),
                    ),
                ),
                "temperature": random(),
                "max_tokens": randint(1, 256),
                "top_p": random(),
            },
            {
                "model": token_hex(8),
                "messages": [
                    {
                        "role": "user",
                        "content": "What is the latest population estimate for {{ location }}?",
                    }
                ],
                "tools": _TOOLS,
                "tool_choice": "required",
                "temperature": random(),
                "max_tokens": randint(1, 256),
                "top_p": random(),
            },
        ],
    )
    def test_round_trip(self, obj: CompletionCreateParamsBase) -> None:
        pv: v1.PromptVersionData = create_prompt_version_from_openai(obj)
        messages, kwargs = to_chat_messages_and_kwargs(pv, formatter=NO_OP_FORMATTER)
        new_obj: CompletionCreateParamsBase = {"messages": messages, **kwargs}  # type: ignore[typeddict-item]
        assert not DeepDiff(obj, new_obj)


class _MockFormatter:
    def format(self, _: str, /, *, variables: Mapping[str, str]) -> str:
        return json.dumps(variables)



from urllib.parse import quote


def encode_path_param(s: str) -> str:
    """
    Quoting function for FastAPI path parameters.

    Encodes a string for use in URL paths, raising ValueError if it
    contains characters that would cause routing issues (/, ?, #).

    Args:
        s (str): The string to encode

    Returns:
        str: The encoded string safe for path parameters

    Raises:
        ValueError: If the string contains /, ?, or # characters
    """
    for char in "/?#":
        if char in s:
            raise ValueError(f"Cannot encode string containing '{char}' for URL path: {s}")
    return quote(s, safe="")



"""Utilities for generating and regenerating OpenTelemetry-compliant span and trace IDs."""

import random
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Sequence, cast

from phoenix.client.__generated__ import v1

if TYPE_CHECKING:
    import pandas as pd

__all__ = ["uniquify_spans", "uniquify_spans_dataframe", "dataframe_to_spans"]

# Source implementation:opentelemetry.sdk.trace.id_generator.RandomIdGenerator

_INVALID_SPAN_ID = 0x0000000000000000
_INVALID_TRACE_ID = 0x00000000000000000000000000000000


def _generate_trace_id() -> str:
    trace_id = random.getrandbits(128)
    while trace_id == _INVALID_TRACE_ID:
        trace_id = random.getrandbits(128)
    return hex(trace_id)[2:].zfill(32)


def _generate_span_id() -> str:
    span_id = random.getrandbits(64)
    while span_id == _INVALID_SPAN_ID:
        span_id = random.getrandbits(64)
    return hex(span_id)[2:].zfill(16)


def uniquify_spans(
    spans: Sequence[v1.Span],
    *,
    in_place: bool = False,
) -> list[v1.Span]:
    """
    Regenerates span and trace IDs for a sequence of Span objects while maintaining parent-child
    relationships. Typically used when creating spans with the client to ensure that the spans
    have unique OpenTelemetry IDs to avoid collisions and guarantee that the spans can be inserted.

    This utility generates new valid OpenTelemetry-compliant span_ids and trace_ids
    for a collection of spans. The parent-child relationships within the span
    collection are preserved by mapping old IDs to new IDs consistently.

    Args:
        spans: A sequence of Span objects to regenerate IDs for.
        in_place: If True, modifies the original spans. If False (default),
            creates deep copies of the spans before modification.

    Returns:
        A list of Span objects with regenerated IDs.
        If in_place=True, returns the modified input. If in_place=False, returns
        a deep copy with modifications.

    Example:
        >>> from phoenix.client import Client
        >>> from phoenix.client.helpers.spans import uniquify_spans
        >>> client = Client()
        >>>
        >>> # Original spans that may have duplicate IDs
        >>> spans = [...]
        >>>
        >>> # Generate new IDs to ensure uniqueness
        >>> new_spans = uniquify_spans(spans)
        >>>
        >>> # Now create the spans with guaranteed unique IDs
        >>> result = client.spans.create_spans(
        ...     project_identifier="my-project",
        ...     spans=new_spans
        ... )
    """
    if in_place:
        mutable_spans = spans if isinstance(spans, list) else list(spans)
    else:
        mutable_spans = list(deepcopy(spans))

    trace_id_mapping: dict[str, str] = {}
    span_id_mapping: dict[str, str] = {}

    generated_trace_ids: set[str] = set()
    generated_span_ids: set[str] = set()

    for span in mutable_spans:
        if span.get("context"):
            old_trace_id = span["context"].get("trace_id", "")
            if old_trace_id and old_trace_id not in trace_id_mapping:
                new_trace_id = _generate_trace_id()
                while new_trace_id in generated_trace_ids:
                    new_trace_id = _generate_trace_id()
                generated_trace_ids.add(new_trace_id)
                trace_id_mapping[old_trace_id] = new_trace_id

            old_span_id = span["context"].get("span_id", "")
            if old_span_id and old_span_id not in span_id_mapping:
                new_span_id = _generate_span_id()
                while new_span_id in generated_span_ids:
                    new_span_id = _generate_span_id()
                generated_span_ids.add(new_span_id)
                span_id_mapping[old_span_id] = new_span_id

    for span in mutable_spans:
        if span.get("context"):
            old_trace_id = span["context"].get("trace_id", "")
            if old_trace_id in trace_id_mapping:
                span["context"]["trace_id"] = trace_id_mapping[old_trace_id]

            old_span_id = span["context"].get("span_id", "")
            if old_span_id in span_id_mapping:
                span["context"]["span_id"] = span_id_mapping[old_span_id]

        old_parent_id = span.get("parent_id")
        if old_parent_id and old_parent_id in span_id_mapping:
            span["parent_id"] = span_id_mapping[old_parent_id]

    return mutable_spans


def uniquify_spans_dataframe(
    df: "pd.DataFrame",
    *,
    in_place: bool = False,
) -> "pd.DataFrame":
    """
    Regenerates span and trace IDs for a pandas DataFrame while maintaining parent-child
    relationships. Typically used when creating spans with the client to ensure that the spans
    have unique OpenTelemetry IDs to avoid collisions and guarantee that the spans can be inserted.

    This utility generates new valid OpenTelemetry-compliant span_ids and trace_ids
    for a DataFrame of spans (typically from get_spans_dataframe). The parent-child
    relationships within the span collection are preserved by mapping old IDs to new IDs
    consistently.

    Args:
        df: A pandas DataFrame (typically from get_spans_dataframe) to regenerate IDs for.
        in_place: If True, modifies the original DataFrame. If False (default),
            creates a deep copy of the DataFrame before modification.

    Returns:
        A DataFrame with regenerated IDs in the index and columns.
        If in_place=True, returns the modified input. If in_place=False, returns
        a deep copy with modifications.

    Example:
        >>> from phoenix.client import Client
        >>> from phoenix.client.helpers.spans import uniquify_spans_dataframe
        >>> client = Client()
        >>>
        >>> # Get spans as DataFrame
        >>> df = client.spans.get_spans_dataframe(
        ...     project_identifier="my-project"
        ... )
        >>>
        >>> # Generate new IDs for the DataFrame
        >>> new_df = uniquify_spans_dataframe(df)
    """
    import pandas as pd

    if not in_place:
        df = df.copy(deep=True)

    trace_id_mapping: dict[str, str] = {}
    span_id_mapping: dict[str, str] = {}

    if "context.trace_id" in df.columns:
        unique_trace_ids = df["context.trace_id"].dropna().unique()  # pyright: ignore
        for old_trace_id in unique_trace_ids:  # pyright: ignore
            old_trace_id_str = str(old_trace_id)
            if old_trace_id_str and old_trace_id_str not in trace_id_mapping:
                trace_id_mapping[old_trace_id_str] = _generate_trace_id()

    span_ids_to_map: set[str] = set()

    # Add span IDs from index - check if index contains span IDs
    index_name = df.index.name  # pyright: ignore
    index_has_span_ids = False
    if index_name is None or index_name == "span_id" or "span" in str(index_name).lower():  # pyright: ignore
        # Assume index contains span IDs
        index_has_span_ids = True
        span_ids_to_map.update(str(idx) for idx in df.index if pd.notna(idx))  # pyright: ignore

    if "context.span_id" in df.columns:
        span_ids_to_map.update(str(sid) for sid in df["context.span_id"].dropna() if pd.notna(sid))  # pyright: ignore

    for old_span_id in span_ids_to_map:
        if old_span_id and old_span_id not in span_id_mapping:
            span_id_mapping[old_span_id] = _generate_span_id()

    if "context.trace_id" in df.columns:
        df["context.trace_id"] = df["context.trace_id"].map(  # pyright: ignore
            lambda x: trace_id_mapping.get(str(x), x) if pd.notna(x) else x  # pyright: ignore
        )

    if "context.span_id" in df.columns:
        df["context.span_id"] = df["context.span_id"].map(  # pyright: ignore
            lambda x: span_id_mapping.get(str(x), x) if pd.notna(x) else x  # pyright: ignore
        )

    if "parent_id" in df.columns:
        df["parent_id"] = df["parent_id"].map(  # pyright: ignore
            lambda x: span_id_mapping.get(str(x), x) if pd.notna(x) else x  # pyright: ignore
        )

    if index_has_span_ids and span_id_mapping:
        new_index_values: list[Any] = []
        for idx in df.index:  # pyright: ignore
            if pd.notna(idx):  # pyright: ignore
                old_idx_str = str(idx)  # pyright: ignore
                new_idx = span_id_mapping.get(old_idx_str, old_idx_str)
                new_index_values.append(new_idx)  # pyright: ignore
            else:
                new_index_values.append(idx)  # pyright: ignore

        new_index = pd.Index(new_index_values, name=index_name)  # pyright: ignore
        df.index = new_index

    return df


def dataframe_to_spans(df: "pd.DataFrame") -> list[v1.Span]:
    """
    Converts a pandas DataFrame (from get_spans_dataframe) back to a list of Span objects.

    This utility reconstructs Span objects from the flattened DataFrame structure
    returned by get_spans_dataframe. It handles the conversion of flattened column
    names (e.g., 'context.span_id') back to nested dictionaries.

    Args:
        df: A pandas DataFrame typically returned by get_spans_dataframe.

    Returns:
        A list of Span objects reconstructed from the DataFrame.

    Example:
        >>> from phoenix.client import Client
        >>> from phoenix.client.helpers.spans import dataframe_to_spans
        >>>
        >>> client = Client()
        >>>
        >>> # Get spans as DataFrame
        >>> df = client.spans.get_spans_dataframe(
        ...     project_identifier="my-project"
        ... )
        >>>
        >>> # Filter or modify the DataFrame
        >>> filtered_df = df[df['span_kind'] == 'LLM']
        >>>
        >>> # Convert back to Span objects
        >>> spans = dataframe_to_spans(filtered_df)
        >>>
        >>> # Now you can use these spans with other APIs
        >>> result = client.spans.create_spans(
        ...     project_identifier="another-project",
        ...     spans=spans
        ... )
    """
    import pandas as pd

    spans: list[v1.Span] = []

    for idx, row in df.iterrows():  # pyright: ignore
        span: dict[str, Any] = {}

        if df.index.name == "span_id" or df.index.name is None:  # pyright: ignore
            span_id = str(idx)
        else:
            span_id = str(row.get("context.span_id", ""))  # pyright: ignore

        # Build context
        context: dict[str, str] = {}
        # pyright: ignore for pandas Series boolean operations
        if "context.trace_id" in row and pd.notna(row["context.trace_id"]):  # pyright: ignore[reportGeneralTypeIssues,reportUnknownMemberType,reportUnknownArgumentType]
            context["trace_id"] = str(row["context.trace_id"])  # pyright: ignore
        if span_id:
            context["span_id"] = span_id

        if context:
            span["context"] = context

        # Copy direct fields
        direct_fields = [
            "name",
            "span_kind",
            "parent_id",
            "status_code",
            "status_message",
            "start_time",
            "end_time",
        ]

        for field in direct_fields:
            # pyright: ignore for pandas Series boolean operations
            if field in row and pd.notna(row[field]):  # pyright: ignore[reportGeneralTypeIssues,reportUnknownMemberType,reportUnknownArgumentType]
                value = row[field]  # pyright: ignore
                if field in ["start_time", "end_time"]:
                    if hasattr(value, "isoformat"):  # pyright: ignore
                        value = value.isoformat()  # pyright: ignore
                    else:
                        value = str(value)  # pyright: ignore
                span[field] = value

        if "events" in row:
            try:
                events = row["events"]  # pyright: ignore
                # Check if events is not null/nan using pandas-safe method
                # pyright: ignore for pandas Series boolean operations
                if events is not None and not pd.isna(events):  # pyright: ignore[reportGeneralTypeIssues,reportUnknownMemberType,reportUnknownArgumentType]
                    if hasattr(events, "__len__"):  # pyright: ignore
                        if len(events) > 0:  # pyright: ignore
                            if isinstance(events, list):
                                span["events"] = events
                            else:
                                span["events"] = list(events)  # pyright: ignore
                    elif events:  # pyright: ignore[reportGeneralTypeIssues]
                        span["events"] = [events]
            except (ValueError, TypeError):
                pass

        # Reconstruct attributes from flattened columns
        attributes: dict[str, Any] = {}
        for col in df.columns:
            if col.startswith("attributes."):
                attr_name = col[len("attributes.") :]
                value = row[col]  # pyright: ignore

                # Handle different data types that might cause issues with pd.notna()
                try:
                    if isinstance(value, (list, tuple)):
                        attributes[attr_name] = value
                    elif hasattr(value, "__len__") and not isinstance(value, str):  # pyright: ignore[reportUnknownArgumentType]
                        if len(value) > 0:  # pyright: ignore
                            attributes[attr_name] = value
                    elif pd.notna(value):  # pyright: ignore[reportGeneralTypeIssues,reportUnknownMemberType,reportUnknownArgumentType]
                        attributes[attr_name] = value
                except (ValueError, TypeError):
                    if value is not None:
                        attributes[attr_name] = value

        if attributes:
            span["attributes"] = attributes

        if "name" not in span:
            span["name"] = "unknown"
        if "span_kind" not in span:
            span["span_kind"] = "UNKNOWN"
        if "status_code" not in span:
            span["status_code"] = "UNSET"

        if span.get("context", {}).get("span_id") and span.get("context", {}).get("trace_id"):
            spans.append(cast(v1.Span, span))

    return spans



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Iterable, Mapping, Optional

import pytest
from deepdiff.diff import DeepDiff
from faker import Faker

from phoenix.client.__generated__ import v1
from phoenix.client.helpers.sdk.anthropic.messages import (
    _MessageConversion,
    _TextContentPartConversion,
    _ToolCallContentPartConversion,
    _ToolConversion,
    _ToolKwargs,
    _ToolKwargsConversion,
    _ToolResultContentPartConversion,
)
from phoenix.client.utils.template_formatters import NO_OP_FORMATTER

if TYPE_CHECKING:
    from anthropic.types import (
        MessageParam,
        TextBlockParam,
        ToolParam,
        ToolResultBlockParam,
        ToolUseBlockParam,
    )

fake = Faker()


def _dict() -> dict[str, Any]:
    return fake.pydict(3, value_types=(int, float, bool, str))  # pyright: ignore[reportUnknownMemberType]


def _str() -> str:
    return fake.pystr(8, 8)


def _text() -> TextBlockParam:
    return {
        "type": "text",
        "text": _str(),
    }


def _tool_use() -> ToolUseBlockParam:
    return {
        "type": "tool_use",
        "id": _str(),
        "input": json.dumps(_dict()),
        "name": _str(),
    }


def _tool_result() -> ToolResultBlockParam:
    return {
        "type": "tool_result",
        "tool_use_id": _str(),
        "content": _str(),  # TODO: relax this
    }


def _tool(name: Optional[str] = None) -> ToolParam:
    return {
        "name": name or _str(),
        "description": _str(),
        "input_schema": {
            "type": "object",
            "properties": {
                "x": {"type": "int", "description": _str()},
                "y": {"type": "string", "description": _str()},
            },
            "required": ["x", "y"],
            "additionalProperties": False,
        },
    }


class TestMessageConversion:
    @pytest.mark.parametrize(
        "obj",
        [
            {"role": "user", "content": [_text(), _text()]},
            {"role": "assistant", "content": [_text(), _tool_use(), _tool_use()]},
            {"role": "user", "content": [_text(), _tool_result(), _tool_result()]},
            {"role": "assistant", "content": [_text(), _text()]},
            {"role": "user", "content": _str()},
            {"role": "assistant", "content": _str()},
        ],
    )
    def test_round_trip(self, obj: MessageParam) -> None:
        x: v1.PromptMessage = _MessageConversion.from_anthropic(obj)
        new_obj = next(_MessageConversion.to_anthropic(x, {}, NO_OP_FORMATTER))
        assert not DeepDiff(obj, new_obj)


class TestToolConversion:
    @pytest.mark.parametrize(
        "tools",
        [[_tool() for _ in range(3)]],
    )
    def test_round_trip(self, tools: Iterable[ToolParam]) -> None:
        new_tools = list(_ToolConversion.to_anthropic(_ToolConversion.from_anthropic(tools)))
        assert not DeepDiff(list(tools), new_tools)


class TestTextConversion:
    def test_round_trip(self) -> None:
        obj: TextBlockParam = _text()
        x: v1.TextContentPart = _TextContentPartConversion.from_anthropic(obj)
        new_obj: TextBlockParam = _TextContentPartConversion.to_anthropic(x, {}, NO_OP_FORMATTER)
        assert not DeepDiff(obj, new_obj)

    def test_formatter(self) -> None:
        x = v1.TextContentPart(type="text", text=_str())
        formatter, variables = _MockFormatter(), _dict()
        ans: TextBlockParam = _TextContentPartConversion.to_anthropic(x, variables, formatter)
        assert ans["text"] == formatter.format(x["text"], variables=variables)


class TestToolCallConversion:
    def test_round_trip(self) -> None:
        obj: ToolUseBlockParam = _tool_use()
        x: v1.ToolCallContentPart = _ToolCallContentPartConversion.from_anthropic(obj)
        new_obj: ToolUseBlockParam = _ToolCallContentPartConversion.to_anthropic(
            x, {}, NO_OP_FORMATTER
        )
        assert not DeepDiff(obj, new_obj)


class TestToolResultBlockParam:
    def test_round_trip(self) -> None:
        obj: ToolResultBlockParam = _tool_result()
        x: v1.ToolResultContentPart = _ToolResultContentPartConversion.from_anthropic(obj)
        new_obj: ToolResultBlockParam = _ToolResultContentPartConversion.to_anthropic(
            x, {}, NO_OP_FORMATTER
        )
        assert not DeepDiff(obj, new_obj)


class TestToolKwargs:
    @pytest.mark.parametrize(
        "obj",
        [
            {},
            {
                "tools": [_tool(), _tool()],
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": {"type": "auto"},
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": {
                    "type": "auto",
                    "disable_parallel_tool_use": True,
                },
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": {"type": "any"},
            },
            {
                "tools": [_tool(), _tool()],
                "tool_choice": {
                    "type": "any",
                    "disable_parallel_tool_use": True,
                },
            },
            {
                "tools": [_tool(), _tool("xyz")],
                "tool_choice": {
                    "type": "tool",
                    "name": "xyz",
                },
            },
            {
                "tools": [_tool(), _tool("xyz")],
                "tool_choice": {
                    "type": "tool",
                    "name": "xyz",
                    "disable_parallel_tool_use": True,
                },
            },
        ],
    )
    def test_round_trip(self, obj: _ToolKwargs) -> None:
        x: Optional[v1.PromptTools] = _ToolKwargsConversion.from_anthropic(obj)
        new_obj: _ToolKwargs = _ToolKwargsConversion.to_anthropic(x)
        assert not DeepDiff(obj, new_obj)


class _MockFormatter:
    def format(self, _: str, /, *, variables: Mapping[str, str]) -> str:
        return json.dumps(variables)



import httpx

DEFAULT_TIMEOUT = httpx.Timeout(timeout=600.0, connect=5.0)
DEFAULT_MAX_RETRIES = 2
DEFAULT_CONNECTION_LIMITS = httpx.Limits(max_connections=1000, max_keepalive_connections=100)

ENV_OTEL_EXPORTER_OTLP_ENDPOINT = "OTEL_EXPORTER_OTLP_ENDPOINT"

# Phoenix environment variables
ENV_PHOENIX_PORT = "PHOENIX_PORT"
ENV_PHOENIX_GRPC_PORT = "PHOENIX_GRPC_PORT"
ENV_PHOENIX_HOST = "PHOENIX_HOST"
ENV_PHOENIX_HOST_ROOT_PATH = "PHOENIX_HOST_ROOT_PATH"
ENV_PHOENIX_COLLECTOR_ENDPOINT = "PHOENIX_COLLECTOR_ENDPOINT"
ENV_PHOENIX_CLIENT_HEADERS = "PHOENIX_CLIENT_HEADERS"
ENV_PHOENIX_API_KEY = "PHOENIX_API_KEY"

HOST = "0.0.0.0"
PORT = 6006
